from typing import Any, Dict

from .__AV_Common import *


class VideoFilter(AVFilter):


    def __init__(self, media_backend: AVMediaBackend, filter_handle: str, parameters: dict[str, int | float | str], backend_additional_info: dict[Any, Any]):
        filter_type = AVFilterType.AV_TYPE_VIDEO
        super().__init__(filter_type, media_backend, filter_handle, parameters, backend_additional_info)



import vlc
import os
from typing import Dict, Any, Union, List, Callable
from .__AV_Common import *
from .__AV_Instance import AVMediaInstance
from .__AV_Interface import AVMediaInterface
from .__AV_Player import AVPlayer
from .vlc_audio_filter import VLCAudioFilter

def handle_vlc_error(call_func: Callable) -> Any:
    try:
        return call_func()
    except vlc.VLCException as e:
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"VLC Exception: {str(e)}")
    except AttributeError as e:
        if "vlc" in str(e).lower():
            raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER, f"VLC property not available: {str(e)}")
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"Attribute error: {str(e)}")
    except RuntimeError as e:
        if "media" in str(e).lower():
            raise AVError(AVErrorInfo.FILE_NOTFOUND, f"Media loading failed: {str(e)}")
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"Runtime error: {str(e)}")
    except OSError as e:
        raise AVError(AVErrorInfo.FILE_NOTFOUND, f"File error: {str(e)}")
    except TypeError as e:
        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_TYPE, f"Type error: {str(e)}")
    except ValueError as e:
        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_VALUE, f"Value error: {str(e)}")
    except Exception as e:
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"Unknown error: {str(e)}")

class VLCMediaInterface(AVMediaInterface):
    
    def __init__(self) -> None:
        super().__init__(AVMediaType.AV_TYPE_VIDEO, AVMediaBackend.AV_BACKEND_VLC)
        self.__instances: dict[int, str] = {}
        self.__vlc_instance: vlc.Instance | None = None
        self.__media_player: vlc.MediaPlayer | None = None
        self.__current_id: int | None = None
        self.__applied_filters: Dict[int, VLCAudioFilter] = {}
        self.__is_initialized = False
        self.__stopped = False
        self.__end_reached = False

    def init(self, *args, **kw):
        try:
            vlc_args = kw.get("vlc_args", [])
            
            def create_instance():
                self.__vlc_instance = vlc.Instance(vlc_args)
                self.__media_player = self.__vlc_instance.media_player_new()
                
            handle_vlc_error(create_instance)
            
            self.__is_initialized = True
            self.__stopped = False
            self.__end_reached = False
            
        except Exception as e:
            raise AVError(AVErrorInfo.INITIALIZATION_ERROR, f"Failed to initialize VLC: {str(e)}")

    def free(self):
        if self.__media_player is not None:
            try:
                def stop_and_release():
                    self.__media_player.stop()
                    self.__media_player.release()
                    
                handle_vlc_error(stop_and_release)
                self.__media_player = None
                self.__vlc_instance = None
                self.__current_id = None
                self.__applied_filters.clear()
                self.__is_initialized = False
                self.__stopped = False
                self.__end_reached = False
            except Exception:
                pass

    def _check_initialized(self):
        if not self.__is_initialized or self.__vlc_instance is None or self.__media_player is None:
            raise AVError(AVErrorInfo.UNINITIALIZED, "VLC not initialized")

    def _check_instance(self, id: int):
        self._check_initialized()
        if self.__current_id != id:
            raise AVError(AVErrorInfo.INVALID_HANDLE, f"Invalid instance ID: {id}")

    def get_vlc_instance(self):
        self._check_initialized()
        return self.__vlc_instance

    def get_media_player(self):
        self._check_initialized()
        return self.__media_player

    def load_file(self, id: int, path: str):
        self._check_initialized()

        def loadfile():
            media = self.__vlc_instance.media_new(path)
            self.__media_player.set_media(media)
            
        if path and is_path(path) and os.path.exists(path):
            self.__current_id = id
            self.__instances[id] = path
            self.__stopped = False
            self.__end_reached = False
            handle_vlc_error(loadfile)

    def load_url(self, id: int, url: str):
        self._check_initialized()
        
        def loadurl():
            media = self.__vlc_instance.media_new(url)
            self.__media_player.set_media(media)
        
        if url and is_url(url):
            self.__current_id = id
            self.__instances[id] = url
            self.__stopped = False
            self.__end_reached = False
            handle_vlc_error(loadurl)

    def release(self, id: int):
        if self.__current_id == id:
            self._check_initialized()
            
            def stop():
                self.__media_player.stop()
                
            handle_vlc_error(stop)
            self.__current_id = None
            self.__applied_filters.clear()
            self.__stopped = True
            self.__end_reached = False

    def play(self, id: int):
        self._check_instance(id)
        
        def resume():
            if self.__end_reached:
                media = self.__vlc_instance.media_new(self.__instances[id])
                self.__media_player.set_media(media)
                self.__end_reached = False
                self.__stopped = False
            self.__media_player.play()
        
        self.__stopped = False
        handle_vlc_error(resume)

    def pause(self, id: int):
        self._check_instance(id)
        
        def pause_play():
            self.__media_player.pause()
            
        handle_vlc_error(pause_play)

    def mute(self, id: int):
        self._check_instance(id)
        
        def toggle_mute():
            current_mute = self.__media_player.audio_get_mute()
            self.__media_player.audio_set_mute(not current_mute)
            
        handle_vlc_error(toggle_mute)

    def unmute(self, id: int):
        self._check_instance(id)
        
        def unmute_audio():
            self.__media_player.audio_set_mute(False)
            
        handle_vlc_error(unmute_audio)

    def stop(self, id: int):
        self._check_instance(id)
        
        def halt():
            self.__media_player.stop()
            
        handle_vlc_error(halt)
        self.__stopped = True
        self.__end_reached = False

    def set_volume(self, id: int, offset: float):
        self._check_instance(id)
        
        def set_vol():
            volume = int(max(0, min(100, offset)))
            self.__media_player.audio_set_volume(volume)
            
        handle_vlc_error(set_vol)

    def set_position(self, id: int, offset: int):
        self._check_instance(id)
        
        def seek():
            self.__media_player.set_time(offset * 1000)
            
        handle_vlc_error(seek)

    def set_loop(self, id: int, loop: bool):
        self._check_instance(id)
        
        def set_loop_mode():
            media = self.__media_player.get_media()
            if media:
                if loop:
                    media.add_option(":input-repeat=-1")
                else:
                    media.add_option(":input-repeat=0")
                    
        handle_vlc_error(set_loop_mode)

    def get_length(self, id: int) -> int:
        self._check_instance(id)
        
        def get_duration():
            length = self.__media_player.get_length()
            return length // 1000 if length > 0 else 0
            
        return handle_vlc_error(get_duration)

    def get_position(self, id: int) -> int:
        self._check_instance(id)
        
        def get_time():
            time_ms = self.__media_player.get_time()
            return time_ms // 1000 if time_ms > 0 else 0
            
        return handle_vlc_error(get_time)

    def get_play_state(self, id: int) -> AVPlaybackState:
        self._check_instance(id)
        try:
            def get_state():
                return self.__media_player.get_state()
                
            if self.__stopped:
                return AVPlaybackState.AV_STATE_STOPPED
            
            state = handle_vlc_error(get_state)
            
            if state == vlc.State.Playing:
                return AVPlaybackState.AV_STATE_PLAYING
            elif state == vlc.State.Paused:
                return AVPlaybackState.AV_STATE_PAUSED
            elif state == vlc.State.Stopped:
                return AVPlaybackState.AV_STATE_STOPPED
            elif state == vlc.State.Ended:
                self.__end_reached = True
                return AVPlaybackState.AV_STATE_NOTHING
            else:
                return AVPlaybackState.AV_STATE_NOTHING
                
        except AVError:
            return AVPlaybackState.AV_STATE_NOTHING

    def get_mute_state(self, id: int) -> AVMuteState:
        self._check_instance(id)
        
        def get_mute():
            return self.__media_player.audio_get_mute()
            
        muted = handle_vlc_error(get_mute)
        return AVMuteState.AV_AUDIO_MUTED if muted else AVMuteState.AV_AUDIO_UNMUTED

    def get_volume(self, id: int) -> float:
        self._check_instance(id)
        
        def get_vol():
            return float(self.__media_player.audio_get_volume())
            
        return handle_vlc_error(get_vol)

    def get_loop(self, id: int) -> bool:
        self._check_instance(id)
        return False

    def _rebuild_filter_chain(self):
        filter_strings = []
        
        for filter_obj in self.__applied_filters.values():
            filter_string = filter_obj.construct()
            if filter_string:
                filter_strings.append(filter_string)
        
        return ":".join(filter_strings) if filter_strings else ""

    def apply_filter(self, id: int, filter_id: int, filter_struct: AVFilter):
        self._check_instance(id)
        if not isinstance(filter_struct, VLCAudioFilter):
            raise AVError(AVErrorInfo.INVALID_MEDIA_FILTER, "Filter must be VLCAudioFilter")
        
        self.__applied_filters[filter_id] = filter_struct
        
        filter_chain = self._rebuild_filter_chain()
        
        def set_audio_filter():
            if filter_chain:
                self.__media_player.audio_filter_set(filter_chain)
            else:
                self.__media_player.audio_filter_set("")
                
        handle_vlc_error(set_audio_filter)

    def remove_filter(self, id: int, filter_id: int):
        self._check_instance(id)
        if filter_id in self.__applied_filters:
            del self.__applied_filters[filter_id]
            
            filter_chain = self._rebuild_filter_chain()
            
            def set_audio_filter():
                if filter_chain:
                    self.__media_player.audio_filter_set(filter_chain)
                else:
                    self.__media_player.audio_filter_set("")
                    
            handle_vlc_error(set_audio_filter)

    def set_parameter(self, id: int, filter_id: int, parameter_name: str, value: ParameterValue):
        self._check_instance(id)
        if filter_id in self.__applied_filters:
            filter_struct = self.__applied_filters[filter_id]
            filter_struct.set_parameter(parameter_name, value)
            
            filter_chain = self._rebuild_filter_chain()
            
            def set_audio_filter():
                if filter_chain:
                    self.__media_player.audio_filter_set(filter_chain)
                else:
                    self.__media_player.audio_filter_set("")
                    
            handle_vlc_error(set_audio_filter)

    def get_parameter(self, id: int, filter_id: int, parameter_name: str) -> ParameterValue | None:
        self._check_instance(id)
        if filter_id in self.__applied_filters:
            return self.__applied_filters[filter_id].get_parameter(parameter_name)
        return None

    def get_devices(self) -> int:
        self._check_initialized()
        
        def get_device_list():
            devices = self.__media_player.audio_output_device_enum()
            count = 0
            device = devices
            while device:
                count += 1
                device = device.next
            return count
            
        return handle_vlc_error(get_device_list)

    def get_device_info(self, index: int) -> AVDevice:
        self._check_initialized()
        
        def get_device_info_by_index():
            devices = self.__media_player.audio_output_device_enum()
            count = 0
            device = devices
            while device and count < index:
                device = device.next
                count += 1
            if device and count == index:
                return AVDevice(AVMediaBackend.AV_BACKEND_VLC, device.description or device.device or "Unknown")
            raise IndexError("Device index out of range")
            
        return handle_vlc_error(get_device_info_by_index)

    def set_device(self, index: int):
        self._check_initialized()
        
        def set_audio_device():
            devices = self.__media_player.audio_output_device_enum()
            count = 0
            device = devices
            while device and count < index:
                device = device.next
                count += 1
            if device and count == index:
                self.__media_player.audio_output_device_set(device.device)
            else:
                raise IndexError("Device index out of range")
                
        handle_vlc_error(set_audio_device)

    def get_current_device(self) -> int:
        self._check_initialized()
        
        def get_current_device_index():
            current_device = self.__media_player.audio_output_device_get()
            devices = self.__media_player.audio_output_device_enum()
            count = 0
            device = devices
            while device:
                if device.device == current_device:
                    return count
                device = device.next
                count += 1
            return 0
            
        return handle_vlc_error(get_current_device_index)


class VLCVideoPlayer(AVPlayer):
    
    def __init__(self) -> None:
        self.__vlc_interface: VLCMediaInterface = VLCMediaInterface()
        super().__init__(AVMediaType.AV_TYPE_VIDEO, AVMediaBackend.AV_BACKEND_VLC, self.__vlc_interface)

    def init(self, *args, **kw):
        self._controler.init(*args, **kw)

    def release(self):
        if self._primary_instance is not None:
            self._primary_instance.release()
            self._primary_instance = None
        self._controler.free()

    def create_file_instance(self, file_path: str) -> AVMediaInstance:
        if self._primary_instance is None:
            self._primary_instance = AVMediaInstance(self._controler)
        self._primary_instance.load_file(file_path)
        return self._primary_instance

    def create_url_instance(self, url: str) -> AVMediaInstance:
        if self._primary_instance is None:
            self._primary_instance = AVMediaInstance(self._controler)
        self._primary_instance.load_url(url)
        return self._primary_instance

    def set_window(self, window):
        def set_hwnd():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                media_player.set_hwnd(int(window))
                
        handle_vlc_error(set_hwnd)

    def forward(self, offset):
        def seek_forward():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                current_time = media_player.get_time()
                media_player.set_time(current_time + (offset * 1000))
                
        handle_vlc_error(seek_forward)

    def backward(self, offset):
        def seek_backward():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                current_time = media_player.get_time()
                media_player.set_time(max(0, current_time - (offset * 1000)))
                
        handle_vlc_error(seek_backward)

    def set_volume_relative(self, direction, offset):
        def adjust_volume():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                current_volume = media_player.audio_get_volume()
                if direction == "up":
                    new_volume = min(100, current_volume + offset)
                elif direction == "down":
                    new_volume = max(0, current_volume - offset)
                else:
                    return
                media_player.audio_set_volume(int(new_volume))
                
        handle_vlc_error(adjust_volume)

    def set_fullscreen(self, state):
        def toggle_fullscreen():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                media_player.set_fullscreen(state)
                
        handle_vlc_error(toggle_fullscreen)

    def get_fullscreen(self):
        def get_fs_state():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                return media_player.get_fullscreen()
            return False
            
        return handle_vlc_error(get_fs_state)

    def set_playback_speed(self, speed):
        def set_rate():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                media_player.set_rate(speed)
                
        handle_vlc_error(set_rate)

    def set_resolution(self, width, height):
        def set_video_filter():
            media_player = self.__vlc_interface.get_media_player()
            if media_player:
                media_player.video_filter_set(f"scale{{width={width},height={height}}}")
                
        handle_vlc_error(set_video_filter)


from enum import Enum, auto
from typing import Union, Any
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import urlparse
from validators import url, ValidationError


ParameterValue = Union[int, float, bool, str]
formats:dict[str, list[str]] = {
       "audio": ["mp1", "mp2", "mp3", "ogg", "wav", "m4a", "aac", "flac", "wma", "aif", "wv"],
       "video": ["flv", "rm", "3gp", "mp4", "mkv", "mov", "wmv", "mpeg", "avi", "webm"]
}

format_descriptions: dict[str, dict[str, str]] = {
    "audio": {
        "mp1": "MPEG Audio Layer 1",
        "mp2": "MPEG Audio Layer 2",
        "mp3": "MPEG Audio Layer 3",
        "ogg": "Ogg Vorbis Audio Container",
        "wav": "Waveform Audio File Format",
        "m4a": "MPEG-4 Audio (often AAC)",
        "aac": "Advanced Audio Coding",
        "flac": "Free Lossless Audio Codec",
        "wma": "Windows Media Audio",
        "aif": "Audio Interchange File Format",
        "wv":  "WavPack Lossless Audio",
    },
    "video": {
        "flv": "Flash Video",
        "rm":  "RealMedia",
        "3gp": "3rd Generation Partnership Project ",
        "mp4": "MPEG-4 Part 14 Video Container",
        "mkv": "Matroska Multimedia Container",
        "mov": "Apple QuickTime Movie",
        "wmv": "Windows Media Video",
        "mpeg": "Moving Picture Experts Group Video (MPEG-1/MPEG-2)",
        "avi": "Audio Video Interleave",
        "webm": "WebM Video Format (VP8/VP9/AV1 Video, Vorbis/Opus Audio)"
    }
}

class AVErrorInfo(Enum):
      UNKNOWN_ERROR = -1
      INITIALIZATION_ERROR = 0
      FILE_NOTFOUND = 1
      UNSUPPORTED_FORMAT = 2
      INVALID_HANDLE = 3
      HTTP_ERROR = 4
      INVALID_MEDIA_FILTER = 5
      INVALID_FILTER_PARAMETER = 6
      INVALID_FILTER_PARAMETER_TYPE = 7
      INVALID_FILTER_PARAMETER_VALUE = 8
      INVALID_FILTER_PARAMETER_RANGE = 9
      UNINITIALIZED = 10
      NET_ERROR = 11

class AVPlaybackState(Enum):
    AV_STATE_NOTHING = -1
    AV_STATE_STOPPED = 0
    AV_STATE_PLAYING = 1
    AV_STATE_PAUSED = 2
    AV_STATE_BUFFERING = 3
    AV_STATE_LOADING = 4

class AVPlaylistMode(Enum):
    SEQUENTIAL = auto()
    REPEAT_ALL = auto()
    REPEAT_ONE = auto()
    SHUFFLE = auto()

class AVPlaylistState(Enum):
    STOPPED = auto()
    PLAYING = auto()
    PAUSED = auto()
    FINISHED = auto()

class AVMuteState(Enum):
                  AV_AUDIO_MUTED = -1
                  AV_AUDIO_UNMUTED = 1

class AVMediaType(Enum):
        AV_TYPE_AUDIO = "audio"
        AV_TYPE_VIDEO = "video"

class AVMediaBackend(Enum):
        AV_BACKEND_FMOD = "fmod"
        AV_BACKEND_MPV = "mpv"
        AV_BACKEND_VLC = "vlc"

class AVMediaSource(Enum):
      AV_SRC_FILE = 1
      AV_SRC_URL = 2
      AV_SRC_NOT_SET = 3

@dataclass
class AVDevice:
        media_backend:AVMediaBackend
        name: str


class AVError(Exception):
      

      def __init__(self, info:AVErrorInfo, message:str|None = None) -> None:
            self.info = info
            self.message = message
            super().__init__(message if message is not None else f"Info: {info}")

AVFilterType = AVMediaType

class AVFilter:
    
    def __init__(self, filter_type: AVFilterType, media_backend: AVMediaBackend,
                 filter_handle: str, parameters: dict[str, ParameterValue], backend_additional_info: dict[Any, Any]):
        self.__type = filter_type
        self.__backend:AVMediaBackend = media_backend
        self.__handle:str = filter_handle
        self.__parameters:dict[str, ParameterValue] = parameters
        self.__info:dict[Any, Any] = backend_additional_info


    @property
    def type(self):
           return self.__type

    @property
    def backend(self):
           return self.__backend

    @property
    def handle(self):
           return self.__handle

    @property
    def info(self):
           return self.__info



    def get_parameters(self):
            return self.__parameters

    def set_parameters(self, parameters: dict[str, ParameterValue] = {}):
            self.__parameters.update(parameters)

    def get_parameter(self, name: str) -> ParameterValue | None:
        if self.__parameters.__contains__(name):
           return self.__parameters[name]
        return None

    def set_parameter(self, name: str, value: ParameterValue):
            if name in self.__parameters:
                    self.__parameters[name] = value


def is_url(url_string: str, default_scheme: str = "https") -> bool:
    if not isinstance(url_string, str):
        return False
    original_trimmed = url_string.strip()
    if not original_trimmed:
        return False

    if default_scheme not in ("http", "https"):
        default_scheme = "https"
    url_to_validate = original_trimmed
    parsed = urlparse(original_trimmed)
    if not parsed.scheme:
        if original_trimmed.startswith('//'):
            url_to_validate = f"{default_scheme}:{original_trimmed}"
        else:
            potential_host = parsed.path.split('/')[0]
            if '.' in original_trimmed or 'localhost' in potential_host.lower():
                 url_to_validate = f"{default_scheme}://{original_trimmed}"
            else:
                 return False

    try:
        result = url(url_to_validate)
        return bool(result)
    except ValidationError:
        return False
    except Exception:
        return False
       

def is_path(path_str:str) -> bool:
    if not isinstance(path_str, str) or path_str == "":
        return False
    try:
        path = Path(path_str)
        return True
    except:
          return False



from abc import ABC, abstractmethod
from typing import Any, List
from .__AV_Common import *


class AVMediaInterface(ABC):


    def __init__(self, media_type: AVMediaType, media_backend: AVMediaBackend) -> None:
        super().__init__()
        self.__backend = media_backend
        self.__type = media_type
        self.__devices:List[AVDevice] = []
        self.__instances:dict[int, Any] = {}
        self.__effects:dict[int, Any] = {}


    @property
    def backend(self):
        return self.__backend

    @property
    def type(self):
     return self.__type


    @abstractmethod
    def load_file(self, id:int, path:str):
        pass

    @abstractmethod
    def load_url(self, id:int, url:str):
        pass

    @abstractmethod
    def init(self, *args, **kw):
        pass

    @abstractmethod
    def free(self):
        pass

    @abstractmethod
    def release(self, id:int):
        pass

    # Playback
    @abstractmethod
    def play(self, id:int):
        pass

    @abstractmethod
    def pause(self, id:int):
        pass

    @abstractmethod
    def mute(self, id:int):
        pass

    @abstractmethod
    def unmute(self, id:int):
        pass

    @abstractmethod
    def stop(self, id:int):
        pass


    @abstractmethod
    def set_volume(self, id:int, offset:float):
        pass

    @abstractmethod
    def set_position(self, id:int, offset:int):
        pass

    @abstractmethod
    def set_loop(self, id:int, loop:bool):
        pass


        # Geters
    @abstractmethod
    def get_length(self, id:int) -> int:
        pass

    @abstractmethod
    def get_position(self, id:int) -> int:
        pass

    @abstractmethod
    def get_play_state(self, id:int) -> AVPlaybackState:
        pass

    @abstractmethod
    def get_mute_state(self, id:int) -> AVMuteState:
        pass

    @abstractmethod
    def get_volume(self, id:int) -> float:
        pass

    @abstractmethod
    def get_loop(self, id:int) -> bool:
        return False


        # Effects/filters
    @abstractmethod
    def apply_filter(self, id:int, filter_id:int, filter_struct:AVFilter):
        pass

    @abstractmethod
    def remove_filter(self, id:int, filter_id):
        pass

    @abstractmethod
    def set_parameter(self, id:int, filter_id:int, parameter_name:str, value:ParameterValue):
        pass

    @abstractmethod
    def get_parameter(self, id:int, filter_id:int, parameter_name:str) -> ParameterValue |None:
        pass

    # Device handling
    @abstractmethod
    def get_devices(self) -> int:
        pass

    @abstractmethod
    def get_device_info(self, index:int) -> AVDevice:
        pass

    @abstractmethod
    def set_device(self, index:int):
        pass

    @abstractmethod
    def get_current_device(self) -> int:
        pass



import os
import platform
from pathlib import Path
from typing import Union, List, Optional
from ctypes.util import find_library

def find_lib_path(paths: Union[List[str], str], libname: str) -> Optional[str]:
    system = platform.system().lower()
    extensions = {
        'windows': ['.dll'],
        'darwin': ['.dylib', '.so'],
        'linux': ['.so']
    }.get(system, ['.so'])
    
    if isinstance(paths, str):
        paths = [paths]
    
    system_path = find_library(libname)
    if system_path and os.path.exists(system_path):
        return system_path
    
    for base_path in paths:
        path_obj = Path(base_path)
        if path_obj.is_file() and path_obj.name.startswith(libname):
            return str(path_obj)
        
        if path_obj.is_dir():
            for ext in extensions:
                candidates = [
                    f"{libname}{ext}",
                    f"lib{libname}{ext}",
                    f"{libname}.0{ext}",
                    f"lib{libname}.0{ext}"
                ]
                for candidate in candidates:
                    lib_path = path_obj / candidate
                    if lib_path.exists():
                        return str(lib_path)
    
    return None


import pyfmodex as fmod

from typing import Any, Dict
from pyfmodex import enums as fmod_enums

from .__AV_Common import *
from .audio_filter import AudioFilter

class FMODAudioFilter(AudioFilter):


    def __init__(self, filter_handle: str, parameters: dict[str, int | float | str], backend_additional_info: dict[Any, Any]):
        media_backend: AVMediaBackend = AVMediaBackend.AV_BACKEND_FMOD
        super().__init__(media_backend, filter_handle, parameters, backend_additional_info)


class FMODEchoFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {"delay_ms": 500.0, "feedback_percent": 50.0, "dry_level_db": 0.0, "wet_level_db": 0.0}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.ECHO,
            "fmod_param_map": {
                "delay_ms": (fmod_enums.DSP_ECHO.DELAY, float, (1.0, 5000.0, 1.0, 500.0)),
                "feedback_percent": (fmod_enums.DSP_ECHO.FEEDBACK, float, (0.0, 100.0, 0.0, 50.0)),
                "dry_level_db": (fmod_enums.DSP_ECHO.DRYLEVEL, float, (-80.0, 10.0, -80.0, 0.0)),
                "wet_level_db": (fmod_enums.DSP_ECHO.WETLEVEL, float, (-80.0, 10.0, -80.0, 0.0)),
            }}
        super().__init__("ECHO", parameters, backend_additional_info)

class FMODReverbFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {
            "decay_time_ms": 1500.0, "early_delay_ms": 7.0, "late_delay_ms": 11.0,
            "hf_reference_hz": 5000.0, "hf_decay_ratio_percent": 83.0, "diffusion_percent": 100.0,
            "density_percent": 100.0, "low_shelf_frequency_hz": 250.0, "low_shelf_gain_db": 0.0,
            "high_cut_hz": 14500.0, "early_late_mix_percent": 96.0, "wet_level_db": -8.0, "dry_level_db": 0.0,
        }
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.SFXREVERB,
            "fmod_param_map": {
                "decay_time_ms": (fmod_enums.DSP_SFXREVERB.DECAYTIME, float, (100.0, 20000.0, 100.0, 1500.0)),
                "early_delay_ms": (fmod_enums.DSP_SFXREVERB.EARLYDELAY, float, (0.0, 300.0, 0.0, 7.0)),
                "late_delay_ms": (fmod_enums.DSP_SFXREVERB.LATEDELAY, float, (0.0, 100.0, 0.0, 11.0)),
                "hf_reference_hz": (fmod_enums.DSP_SFXREVERB.HFREFERENCE, float, (20.0, 20000.0, 20.0, 5000.0)),
                "hf_decay_ratio_percent": (fmod_enums.DSP_SFXREVERB.HFDECAYRATIO, float, (10.0, 100.0, 10.0, 83.0)),
                "diffusion_percent": (fmod_enums.DSP_SFXREVERB.DIFFUSION, float, (0.0, 100.0, 0.0, 100.0)),
                "density_percent": (fmod_enums.DSP_SFXREVERB.DENSITY, float, (0.0, 100.0, 0.0, 100.0)),
                "low_shelf_frequency_hz": (fmod_enums.DSP_SFXREVERB.LOWSHELFFREQUENCY, float, (20.0, 1000.0, 20.0, 250.0)),
                "low_shelf_gain_db": (fmod_enums.DSP_SFXREVERB.LOWSHELFGAIN, float, (-36.0, 12.0, -36.0, 0.0)),
                "high_cut_hz": (fmod_enums.DSP_SFXREVERB.HIGHCUT, float, (20.0, 20000.0, 20.0, 14500.0)),
                "early_late_mix_percent": (fmod_enums.DSP_SFXREVERB.EARLYLATEMIX, float, (0.0, 100.0, 0.0, 96.0)),
                "wet_level_db": (fmod_enums.DSP_SFXREVERB.WETLEVEL, float, (-80.0, 20.0, -80.0, -8.0)),
                "dry_level_db": (fmod_enums.DSP_SFXREVERB.DRYLEVEL, float, (-80.0, 20.0, -80.0, 0.0)),
            }}
        super().__init__("REVERB", parameters, backend_additional_info)

class FMODLowpassFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {"cutoff_frequency_hz": 5000.0, "q_factor": 1.0}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.MULTIBAND_EQ,
            "fixed_param_vals": {
                "filter_type": (fmod_enums.DSP_MULTIBAND_EQ.A_FILTER, fmod_enums.DSP_MULTIBAND_EQ_FILTER_TYPE.LOWPASS_24DB, int),
            },
            "fmod_param_map": {
                "cutoff_frequency_hz": (fmod_enums.DSP_MULTIBAND_EQ.A_FREQUENCY, float, (20.0, 22000.0, 1.0, 5000.0)),
                "q_factor": (fmod_enums.DSP_MULTIBAND_EQ.A_Q, float, (0.1, 10.0, 0.1, 1.0)),
            }}
        super().__init__("LOWPASS", parameters, backend_additional_info)

class FMODHighpassFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {"cutoff_frequency_hz": 250.0, "q_factor": 1.0}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.MULTIBAND_EQ,
            "fixed_param_vals": {
                "filter_type": (fmod_enums.DSP_MULTIBAND_EQ.A_FILTER, fmod_enums.DSP_MULTIBAND_EQ_FILTER_TYPE.HIGHPASS_24DB, int),
            },
            "fmod_param_map": {
                "cutoff_frequency_hz": (fmod_enums.DSP_MULTIBAND_EQ.A_FREQUENCY, float, (20.0, 22000.0, 1.0, 250.0)),
                "q_factor": (fmod_enums.DSP_MULTIBAND_EQ.A_Q, float, (0.1, 10.0, 0.1, 1.0)),
            }}
        super().__init__("HIGHPASS", parameters, backend_additional_info)

class FMODBandpassFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {"center_frequency_hz": 1000.0, "bandwidth_q": 1.0}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.MULTIBAND_EQ,
            "fixed_param_vals": {
                "filter_type": (fmod_enums.DSP_MULTIBAND_EQ.A_FILTER, fmod_enums.DSP_MULTIBAND_EQ_FILTER_TYPE.BANDPASS, int),
            },
            "fmod_param_map": {
                "center_frequency_hz": (fmod_enums.DSP_MULTIBAND_EQ.A_FREQUENCY, float, (20.0, 22000.0, 1.0, 1000.0)),
                "bandwidth_q": (fmod_enums.DSP_MULTIBAND_EQ.A_Q, float, (0.1, 10.0, 0.1, 1.0)),
            }}
        super().__init__("BANDPASS", parameters, backend_additional_info)

class FMODChorusFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {"mix_percent": 50.0, "rate_hz": 0.8, "depth_percent": 3.0}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.CHORUS,
            "fmod_param_map": {
                "mix_percent": (fmod_enums.DSP_CHORUS.MIX, float, (0.0, 100.0, 0.0, 50.0)),
                "rate_hz": (fmod_enums.DSP_CHORUS.RATE, float, (0.0, 20.0, 0.0, 0.8)),
                "depth_percent": (fmod_enums.DSP_CHORUS.DEPTH, float, (0.0, 100.0, 0.0, 3.0)),
            }}
        super().__init__("CHORUS", parameters, backend_additional_info)

class FMODCompressorFilter(FMODAudioFilter):
    def __init__(self):
        parameters = {
            "threshold_db": 0.0, "ratio": 2.5, "attack_ms": 20.0,
            "release_ms": 100.0, "makeup_gain_db": 0.0,
            "use_sidechain": False, "linked_channels": True,
        }
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.COMPRESSOR,
            "fmod_param_map": {
                "threshold_db": (fmod_enums.DSP_COMPRESSOR.THRESHOLD, float, (-80.0, 0.0, 1.0, 0.0)),
                "ratio": (fmod_enums.DSP_COMPRESSOR.RATIO, float, (1.0, 50.0, 0.1, 2.5)),
                "attack_ms": (fmod_enums.DSP_COMPRESSOR.ATTACK, float, (0.1, 1000.0, 0.1, 20.0)),
                "release_ms": (fmod_enums.DSP_COMPRESSOR.RELEASE, float, (10.0, 5000.0, 1.0, 100.0)),
                "makeup_gain_db": (fmod_enums.DSP_COMPRESSOR.GAINMAKEUP, float, (-30.0, 30.0, 0.1, 0.0)),
                "use_sidechain": (fmod_enums.DSP_COMPRESSOR.USESIDECHAIN, bool, (False, True, 1, False)),
                "linked_channels": (fmod_enums.DSP_COMPRESSOR.LINKED, bool, (False, True, 1, True)),
            }}
        super().__init__("COMPRESSOR", parameters, backend_additional_info)

class FMODFlangerFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {"mix_percent": 50.0, "depth_factor": 1.0, "rate_hz": 0.1}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.FLANGE,
            "fmod_param_map": {
                "mix_percent": (fmod_enums.DSP_FLANGE.MIX, float, (0.0, 100.0, 0.0, 50.0)),
                "depth_factor": (fmod_enums.DSP_FLANGE.DEPTH, float, (0.01, 1.0, 0.01, 1.0)),
                "rate_hz": (fmod_enums.DSP_FLANGE.RATE, float, (0.0, 20.0, 0.0, 0.1)),
            }}
        super().__init__("FLANGER", parameters, backend_additional_info)

class FMODDistortionFilter(FMODAudioFilter):
    def __init__(self):
        parameters:dict[str, ParameterValue] = {"level_factor": 0.5}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.DISTORTION,
            "fmod_param_map": {
                "level_factor": (fmod_enums.DSP_DISTORTION.LEVEL, float, (0.0, 1.0, 0.0, 0.5)),
            }}
        super().__init__("DISTORTION", parameters, backend_additional_info)

class FMODPitchShiftFilter(FMODAudioFilter):
    def __init__(self):
        parameters = {"pitch_scale": 1.0, "fft_size_samples": 1024}
        backend_additional_info = {
            "fmod_dsp_type": fmod_enums.DSP_TYPE.PITCHSHIFT,
            "fmod_param_map": {
                "pitch_scale": (fmod_enums.DSP_PITCHSHIFT.PITCH, float, (0.5, 2.0, 0.01, 1.0)),
                "fft_size_samples": (fmod_enums.DSP_PITCHSHIFT.FFTSIZE, float, (256, 4096, 64, 1024)),
            }}
        super().__init__("PITCH_SHIFT", parameters, backend_additional_info)


import os

from .media_info import MediaInfo
from .playlist import PlaylistManager, PlaylistParser, Playlist, PlaylistFormat
from .__AV_Common import *
from .__AV_Instance import AVMediaInstance
from .__utils import find_lib_path

use_fmod = os.environ.get("USE_FMOD", None)
use_mpv = os.environ.get("USE_MPV", None)
use_vlc = os.environ.get("USE_VLC", None)
fmod_lib_path = os.environ.get("FMOD_LIB_PATH", None)
mpv_lib_path = os.environ.get("MPV_LIB_PATH", None)
vlc_lib_path = os.environ.get("VLC_LIB_PATH", None)

if use_fmod:
    if os.path.exists(fmod_lib_path): # type: ignore
        os.environ["PYFMODEX_DLL_PATH"] = os.path.abspath(find_lib_path(fmod_lib_path, "fmod")) # type: ignore

    try:
        from .fmod_audio_player import FMODAudioPlayer
        from .fmod_audio_filter import (FMODAudioFilter, FMODBandpassFilter, FMODChorusFilter, FMODCompressorFilter,
                                        FMODDistortionFilter, FMODEchoFilter, FMODFlangerFilter,
                                        FMODLowpassFilter, FMODHighpassFilter, FMODPitchShiftFilter, FMODReverbFilter
                                        )
        AudioPlayer = FMODAudioPlayer
    except Exception as e:
        raise RuntimeError(f"Error initializing FMOD.\nDetails: {str(e)}")

if use_mpv:
    if os.path.exists(mpv_lib_path): # type: ignore
        os.environ["PATH"] += os.path.abspath(mpv_lib_path)+os.pathsep # type: ignore
    try:
        from .mpv_video_player import MPVVideoPlayer
        from .mpv_audio_filter import (MPVAudioFilter, MPVBandPassFilter, MPVChorusFilter, MPVCompressorFilter,
                                       MPVEchoFilter, MPVFlangerFilter, MPVGateFilter, MPVHighPassFilter,
                                       MPVLimiterFilter, MPVLowPassFilter, MPVPitchShiftFilter
                                       )
        VideoPlayer = MPVVideoPlayer
    except Exception as e:
        raise RuntimeError(f"Error initializing MPV.\nDetails: {str(e)}")

if use_vlc:
    if os.path.exists(vlc_lib_path): # type: ignore
        os.environ["PATH"] += os.path.abspath(vlc_lib_path)+os.pathsep # type: ignore
    try:
        from .vlc_video_player import VLCVideoPlayer
        from .vlc_audio_filter import (VLCAudioFilter, VLCEqualizerFilter)
        VideoPlayer = VLCVideoPlayer
    except Exception as e:
        raise RuntimeError(f"Error initializing VLC.\nDetails: {str(e)}")


if not use_fmod and not use_mpv and not use_vlc:
    raise Exception(
        "No media backend is initialized or has been detected.\n"
        "Define either of: USE_FMOD = 1 or USE_MPV = 1\n"
        "in the environment. Then set the appropriate paths:\n"
        "FMOD_LIB_PATH or MPV_LIB_PATH\n"
        "Example using FMOD:\n"
        "os.environ[\"USE_FMOD\"] = \"1\"\n"
        "os.environ[\"PATH\"] = \"./lib/\""
    )




import os
import pyfmodex as fmod

from pyfmodex import flags as fmod_flags
from pyfmodex import enums as fmod_enums
from pyfmodex.exceptions import FmodError
from pyfmodex.structobject import Structobject
from pyfmodex.structures import CREATESOUNDEXINFO

from typing import Dict, List, Callable, Any, Tuple, Union
from enum import Enum

from .fmod_audio_filter import FMODAudioFilter
from .__AV_Common import *
from .__AV_Instance import AVMediaInstance
from .__AV_Interface import AVMediaInterface
from .__AV_Player import AVPlayer

MIN_FMOD_VERSION = 0x00020108
RangeType = Union[int,float]
fmod_res = fmod_enums.RESULT
TIMEUNIT = fmod_enums.TIMEUNIT

def handle_call_err(call_func:Callable) -> Any:
    try:
        return call_func()
    except FmodError as fmod_err:
        result = fmod_err.result
        info = AVErrorInfo.UNKNOWN_ERROR
        if result == fmod_res.FILE_NOTFOUND:
            info = AVErrorInfo.FILE_NOTFOUND
        elif result == fmod_res.FORMAT:
            info = AVErrorInfo.UNSUPPORTED_FORMAT
        elif result == fmod_res.INVALID_HANDLE or result == fmod_res.CHANNEL_STOLEN:
            info = AVErrorInfo.INVALID_HANDLE
        elif result in [fmod_res.HTTP, fmod_res.HTTP_ACCESS, fmod_res.HTTP_PROXY_AUTH, fmod_res.HTTP_SERVER_ERROR, fmod_res.HTTP_TIMEOUT]:
            info = AVErrorInfo.HTTP_ERROR
        elif result in [fmod_res.NET_CONNECT, fmod_res.NET_SOCKET_ERROR, fmod_res.NET_URL]:
            info = AVErrorInfo.NET_ERROR
        elif result == fmod_res.UNINITIALIZED:
            info = AVErrorInfo.UNINITIALIZED
        else:
            info = AVErrorInfo.UNKNOWN_ERROR
        raise AVError(info, f"Details: {fmod_err.result}")
    except Exception as e:
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"unknown error: {str(e)}")


def safe_assert(condition) -> bool:
    """Custom assert that returns True/False instead of raising AssertionError"""
    return bool(condition)

class AudioMediaInterface(AVMediaInterface):


    def __init__(self) -> None:
        super().__init__(AVMediaType.AV_TYPE_AUDIO, AVMediaBackend.AV_BACKEND_FMOD)
        self.__instances:Dict[int, fmod.sound.Sound] = {}
        self.__effects:Dict[int, fmod.dsp.DSP] = {}
        self.__effect_structs:Dict[int, AVFilter] = {}
        self.__instance_effects:Dict[int, List[int]] = {}
        self.__channels:Dict[int, fmod.channel.Channel] = {}
        self.__system = fmod.System()

        if self.__system.version < MIN_FMOD_VERSION:
            raise Exception(f"Error: Detected FMOD version is lower then the mimum supported version.\nMinimum version is {MIN_FMOD_VERSION}")


    def init(self, *args, **kw):
        max_channels = kw["max_channels"] if "max_channels" in kw else 64
        flags = fmod_flags.INIT_FLAGS.NORMAL | fmod_flags.INIT_FLAGS.PROFILE_ENABLE
        try:
            self.__system.init(maxchannels=max_channels, flags=flags)
            self.__system.stream_buffer_size = Structobject(size=64 * 1024, unit=fmod_enums.TIMEUNIT.RAWBYTES)
        except Exception as e:
            raise AVError(AVErrorInfo.INITIALIZATION_ERROR, f"Error initializing FMOD. {str(e)}")

    def free(self):
        for instance in list(self.__instances.keys()):
            try:
                self.release(instance)
            except Exception:
               pass

        for instance in list(self.__instance_effects.keys()):
            for effect in self.__instance_effects[instance].copy():
                try:
                    self.remove_filter(instance, effect)
                except:
                    pass

        self.__effects.clear()
        self.__instance_effects.clear()
        self.__effect_structs.clear()
        try:
            self.__system.release()
        except Exception as e:
            raise AVError(AVErrorInfo.INVALID_HANDLE)

    def load_file(self, id:int, path:str):
        try:
            self.release(id)
        except: 
            pass
        def create_instance():
            if (path is not None and path != "") and (id not in self.__instances or self.__instances[id] is None):
                self.__instances[id] = self.__system.create_stream(path, mode=fmod_flags.MODE.NONBLOCKING)

        handle_call_err(create_instance)

    def load_url(self, id:int, url:str):
        def create_instance():
            MODE = fmod_flags.MODE
            exinfo = CREATESOUNDEXINFO(filebuffersize=1024 * 16)
            self.__instances[id] = self.__system.create_sound(url, mode=MODE.CREATESTREAM | MODE.NONBLOCKING, exinfo=exinfo)

        if (url is not None and url != "") and (id not in self.__instances or self.__instances[id] is None):
            handle_call_err(create_instance)

    def release(self, id:int):
        try:
            self.stop(id)
        except:
            pass

        if id in self.__channels:
            del self.__channels[id]
            
        if id in self.__instances and self.__instances[id] is not None:
            try:
                handle_call_err(lambda: self.__instances[id].release())
            except:
                pass
            del self.__instances[id]


    def play(self, id:int):
        self.__check_handle(id)

        fmod_state = fmod_enums.OPENSTATE
        sound:fmod.sound.Sound = self.__instances[id]
        state:fmod_enums.OPENSTATE|None = None

        def create_channel(): self.__channels[id] = sound.play(paused=True)
        
        def play_channel():
            starving:bool|None = getattr(sound.open_state, "starving", None)
            if self.__channels[id] and starving is not None:
                self.__channels[id].mute = starving

                self.__channels[id].paused = False
                if id in self.__instance_effects and self.__instance_effects[id] is not None:
                    for effect in self.__instance_effects[id]:
                        self.apply_filter(id, effect, self.__effect_structs[effect])

        try:
            state = getattr(sound.open_state, "state")
            if state == fmod_state.BUFFERING or state == fmod_state.LOADING or state == fmod_state.CONNECTING:
                pass
        except Exception:
            pass

        def play_sound():
            if state == fmod_state.READY or state == fmod_state.PLAYING:
                try:
                    if id not in self.__channels or self.__channels[id] is None:
                        create_channel()
                    play_channel()
                except FmodError as fmod_err:
                    if fmod_err.result == fmod_res.INVALID_HANDLE or fmod_err.result == fmod_res.CHANNEL_STOLEN:
                        create_channel()
                    play_channel()

        handle_call_err(play_sound)

    def pause(self, id:int):
        self.__check_handle(id)

        def pause_channel():
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return
            channel.paused = True  # type: ignore

        try: handle_call_err(pause_channel)
        except: pass

    def mute(self, id:int):
        self.__check_handle(id)

        def mute_channel():
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return
            channel.mute = True  # type: ignore

        try: handle_call_err(mute_channel)
        except: pass

    def unmute(self, id:int):
        self.__check_handle(id)

        def unmute_channel():
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return
            channel.mute = False  # type: ignore

        try: handle_call_err(unmute_channel)
        except: pass

    def stop(self, id:int):
        self.__check_handle(id)

        def stop_channel():
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return
            channel.stop()  # type: ignore

        try: handle_call_err(stop_channel)
        except: pass

    def set_volume(self, id:int, offset:float):
        self.__check_handle(id)

        def set_channel_volume():
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return
            channel.volume = offset  # type: ignore

        try: handle_call_err(set_channel_volume)
        except: pass

    def set_position(self, id:int, offset:int):
        self.__check_handle(id)

        def set_channel_pos():
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return
            channel.set_position(pos=offset*1000, unit=TIMEUNIT.MS)  # type: ignore

        try: handle_call_err(set_channel_pos)
        except: pass

    def set_loop(self, id:int, loop:bool):
        sound:fmod.sound.Sound|None  = self.__check_handle(id)

        def set_loop_mode():
            MODE = fmod_flags.MODE
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(sound is not None) or not safe_assert(channel is not None):
                return
            current_pos:int = channel.get_position(unit=TIMEUNIT.MS)  # type: ignore
            sound.mode = sound.mode | MODE.LOOP_NORMAL if loop else sound.mode & ~MODE.LOOP_NORMAL | MODE.LOOP_OFF  # type: ignore
            channel.mode = channel.mode | MODE.LOOP_NORMAL if loop else channel.mode & ~MODE.LOOP_NORMAL | MODE.LOOP_OFF  # type: ignore
            sound.loop_count = -1 if loop else 0  # type: ignore
            channel.loop_count = -1 if loop else 0  # type: ignore
            if loop: channel.set_position(pos=current_pos, unit=TIMEUNIT.MS)  # type: ignore

        try: handle_call_err(set_loop_mode)
        except: pass

    def get_length(self, id:int) -> int:
        sound:fmod.sound.Sound|None = self.__check_handle(id)

        def get_media_length() -> int:
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(sound is not None) or not safe_assert(channel is not None):
                return -1
            return int(sound.get_length(ltype=TIMEUNIT.MS) /1000)  # type: ignore

        try:
            return handle_call_err(get_media_length)
        except Exception as e:
            return -1

    def get_position(self, id:int) -> int:
        self.__check_handle(id)

        def get_channel_pos() -> int:
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return -1
            return int(channel.get_position(unit=TIMEUNIT.MS) /1000)  # type: ignore

        try:
            return handle_call_err(get_channel_pos)
        except Exception as e:
            return -1

    def get_play_state(self, id:int) -> AVPlaybackState:
        fmod_state = fmod_enums.OPENSTATE
        sound:fmod.sound.Sound|None  = self.__check_handle(id)

        def get_state() -> AVPlaybackState:
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(sound is not None):
                return AVPlaybackState.AV_STATE_NOTHING
            open_state: Structobject| None = None

            try: open_state = sound.open_state  # type: ignore
            except: return AVPlaybackState.AV_STATE_NOTHING
            
            if not safe_assert(open_state is not None):
                return AVPlaybackState.AV_STATE_NOTHING
            state:fmod_enums.OPENSTATE | None = getattr(open_state, "state", None)  # type: ignore

            if channel is None and state == fmod_state.BUFFERING: return AVPlaybackState.AV_STATE_BUFFERING
            elif channel is None and (state == fmod_state.LOADING or state == fmod_state.CONNECTING): return AVPlaybackState.AV_STATE_LOADING
            elif channel is None and state == fmod_state.READY: return AVPlaybackState.AV_STATE_STOPPED

            if not safe_assert(channel is not None):
                return AVPlaybackState.AV_STATE_NOTHING
            if channel.is_playing and channel.paused: return AVPlaybackState.AV_STATE_PAUSED  # type: ignore
            elif channel.is_playing and not channel.paused: return AVPlaybackState.AV_STATE_PLAYING  # type: ignore
            else: return AVPlaybackState.AV_STATE_NOTHING

        try:
            return handle_call_err(get_state)
        except FmodError as fmod_err:
            result = fmod_err.result 
            return AVPlaybackState.AV_STATE_STOPPED if result == fmod_res.CHANNEL_STOLEN or result == fmod_res.INVALID_HANDLE else AVPlaybackState.AV_STATE_NOTHING
        except:
            return AVPlaybackState.AV_STATE_NOTHING

    def get_mute_state(self, id:int) -> AVMuteState:
        MODE = fmod_flags.MODE
        self.__check_handle(id)

        def get_mute_state() -> AVMuteState:
            channel = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return AVMuteState.AV_AUDIO_UNMUTED
            return AVMuteState.AV_AUDIO_MUTED if channel.mute else AVMuteState.AV_AUDIO_UNMUTED  # type: ignore

        try:
            return handle_call_err(get_mute_state)
        except Exception as e:
            return AVMuteState.AV_AUDIO_UNMUTED

    def get_volume(self, id:int) -> float:
        self.__check_handle(id)

        def get_channel_volume() -> float:
            channel:fmod.channel.Channel | None = self.__check_channel(id)
            if not safe_assert(channel is not None):
                return -1
            return channel.volume  # type: ignore

        try:
            return handle_call_err(get_channel_volume)
        except Exception as e:
            return -1

    def get_loop(self, id:int) -> bool:
        MODE = fmod_flags.MODE
        sound:fmod.sound.Sound|None = self.__check_handle(id)

        def get_loop_mode():
            if not safe_assert(sound is not None):
                return False
            return True if bool(sound.mode & MODE.LOOP_OFF) else False  # type: ignore

        try:
            return handle_call_err(get_loop_mode)
        except:
            return False

    def apply_filter(self, id:int, filter_id:int, filter_struct:AVFilter):
        self.__check_handle(id)
        if filter_struct is None or not isinstance(filter_struct, FMODAudioFilter):
            raise AVError(AVErrorInfo.INVALID_MEDIA_FILTER)

        if filter_id not in self.__effect_structs or self.__effect_structs[filter_id] is None:
            self.__effect_structs[filter_id] = filter_struct
        else:
            self.__effect_structs[filter_id].set_parameters(filter_struct.get_parameters())

        if id not in self.__instance_effects or self.__instance_effects[id] is None:
            self.__instance_effects[id] = []
        if filter_id not in self.__instance_effects[id]:
            self.__instance_effects[id].append(filter_id)

        def apply_effect():
            dsp_type:fmod_enums.DSP_TYPE = filter_struct.info["fmod_dsp_type"]
            dsp_obj:fmod.dsp.DSP = self.__system.create_dsp_by_type(dsp_type)
            channel:fmod.channel.Channel|None = self.__check_channel(id)

            if not safe_assert(channel is not None):
                return
            

            self.__effects[filter_id] = dsp_obj
            channel.add_dsp(0, dsp_obj)  # type: ignore

            if "fixed_param_vals" in filter_struct.info:
                for param in filter_struct.info["fixed_param_vals"]:
                    param_data:Tuple[Enum, Enum, ParameterValue] = filter_struct.info["fixed_param_vals"][param]
                    param_type, param_val, param_val_type=param_data
                    if isinstance(param_val_type, int):
                        dsp_obj.set_parameter_int(param_type, param_val)
                    elif isinstance(param_val_type, float):
                        dsp_obj.set_parameter_float(param_type, param_val)
            for param in filter_struct.get_parameters():
                self.__set_parameter_value(dsp_obj, filter_struct, param, filter_struct.get_parameters()[param])


        if filter_id not in self.__effects:
            handle_call_err(apply_effect)

    def remove_filter(self, id:int, filter_id):
        self.__check_handle(id)

        def remove_effect():
            channel:fmod.channel.Channel|None = self.__check_channel(id)
            if filter_id not in self.__effects:
                return
            dsp_obj:fmod.dsp.DSP = self.__effects[filter_id]
            

            if not safe_assert(channel is not None):

                try:
                    dsp_obj.release()
                except:
                    pass
            else:

                try:
                    channel.remove_dsp(dsp_obj)  # type: ignore
                    dsp_obj.release()
                except FmodError as fmod_err:

                    if fmod_err.result == fmod_res.INVALID_HANDLE or fmod_err.result == fmod_res.CHANNEL_STOLEN:
                        try:
                            dsp_obj.release()
                        except:
                            pass
                    else:
                        raise
            

            del self.__effects[filter_id]
            if filter_id in self.__effect_structs: 
                self.__effect_structs.pop(filter_id)
            if id in self.__instance_effects and filter_id in self.__instance_effects[id]: 
                self.__instance_effects[id].remove(filter_id)

        try:
            handle_call_err(remove_effect)
        except AVError as av_err:

            if "INVALID_HANDLE" not in str(av_err):
                raise

    def set_parameter(self, id:int, filter_id:int, parameter_name:str, value:ParameterValue):
        self.__check_handle(id)

        def set_param():
            dsp_obj:fmod.dsp.DSP = self.__check_dsp(filter_id)
            if filter_id not in self.__effect_structs or self.__effect_structs[filter_id]is None:
                raise AVError(AVErrorInfo.INVALID_HANDLE)
            self.__set_parameter_value(dsp_obj, self.__effect_structs[filter_id], parameter_name, value)

        handle_call_err(set_param)

    def get_parameter(self, id:int, filter_id:int, parameter_name:str) -> ParameterValue |None:
        self.__check_handle(id)

        def get_param():
            if filter_id not in self.__effect_structs or self.__effect_structs[filter_id]is None:
                raise AVError(AVErrorInfo.INVALID_HANDLE)
            return self.__effect_structs[filter_id].get_parameters()[parameter_name]

        return handle_call_err(get_param)

    def __check_handle(self, id:int) -> fmod.sound.Sound | None:
        if id not in self.__instances or self.__instances[id] is None or not self.__instances[id]._ptr:
            raise AVError(AVErrorInfo.INVALID_HANDLE)
        return self.__instances[id]

    def __check_channel(self, id:int) -> fmod.channel.Channel | None:
        if id not in self.__channels or self.__channels[id] is None:
            return None
        return self.__channels[id]

    def __check_dsp(self, dsp_id:int) -> fmod.dsp.DSP:
        if dsp_id not in self.__effects or self.__effects[dsp_id] is None:
            raise AVError(AVErrorInfo.INVALID_HANDLE)
        return self.__effects[dsp_id]

    def __set_parameter_value(self, dsp_obj:fmod.dsp.DSP, effect_struct:AVFilter, parameter_name:str, parameter_value:ParameterValue):
        if parameter_name not in effect_struct.info["fmod_param_map"]:
            raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER)

        param_info:Tuple[Enum, type, Tuple[RangeType, RangeType, RangeType, RangeType]] = effect_struct.info["fmod_param_map"][parameter_name]
        dsp_param_type:Enum = param_info[0]
        param_type:type = param_info[1]
        param_range:Tuple[RangeType, RangeType, RangeType, RangeType] = param_info[2]
        param_val:ParameterValue = parameter_value


        if not isinstance(param_val, param_type):
            raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_TYPE)


        if isinstance(param_val, (float, int)) and not (param_val >= param_range[0] and param_val <= param_range[1]):
            raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_RANGE)

        if param_type is int:
            dsp_obj.set_parameter_int(dsp_param_type, param_val)
            effect_struct.set_parameter(parameter_name, param_val)
        elif param_type is float:
            dsp_obj.set_parameter_float(dsp_param_type, param_val)
            effect_struct.set_parameter(parameter_name, param_val)
        elif param_type is bool:
            dsp_obj.set_parameter_bool(dsp_param_type, param_val)
            effect_struct.set_parameter(parameter_name, param_val)
        else:
            raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_VALUE)


    def get_devices(self) -> int:
        def get_device_count() -> int:
            return self.__system.num_drivers

        return handle_call_err(get_device_count)

    def get_device_info(self, index:int) -> AVDevice:
        def device_info() -> AVDevice:
            device:Structobject = self.__system.get_driver_info(index)
            name:str = getattr(device, "name", "none")
            return AVDevice(AVMediaBackend.AV_BACKEND_FMOD, name)

        return handle_call_err(device_info)

    def set_device(self, index:int):
        def set_device_index():
            self.__system.driver = index

        handle_call_err(set_device_index)

    def get_current_device(self) -> int:
        def get_device() -> int:
            return self.__system.driver

        return handle_call_err(get_device)


class FMODAudioPlayer(AVPlayer):


    def __init__(self) -> None:
        self.__media_type = AVMediaType.AV_TYPE_AUDIO
        self.__media_backend = AVMediaBackend.AV_BACKEND_FMOD
        self._controler:AudioMediaInterface = AudioMediaInterface()
        super().__init__(self.__media_type, self.__media_backend, self.__controler)

    def init(self, *args, **kw):
        self._controler.init()

    def release(self):
        super().release()
        self._controler.free()




from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Dict, Optional, Union, Type, Any, Callable, Iterable
from pathlib import Path

import os
import json
import re
import urllib.parse
import urllib.request


class PlaylistFormat(Enum):
    M3U = auto()
    M3U8 = auto()
    XSPF = auto()
    PLS = auto()
    JSON = auto()
    UNKNOWN = auto()


@dataclass
class PlaylistEntry:
    location: str
    title: Optional[str] = None
    duration: Optional[int] = None
    artist: Optional[str] = None
    album: Optional[str] = None
    metadata: Dict[str, Any]|None = None
    
    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}


class PlaylistParser(ABC):
    @classmethod
    @abstractmethod
    def parse(cls, content: str) -> List[PlaylistEntry]:
        pass
    
    @classmethod
    @abstractmethod
    def serialize(cls, entries: List[PlaylistEntry], original_content: Optional[str] = None) -> str:
        pass
    
    @staticmethod
    def normalize_path(path: str) -> str:
        if re.match(r'^https?://', path):
            return path
        
        return os.path.normpath(path)
    
    @staticmethod
    def is_url(path: str) -> bool:
        return bool(re.match(r'^https?://', path))


class M3UParser(PlaylistParser):
    @classmethod
    def parse(cls, content: str) -> List[PlaylistEntry]:
        entries = []
        lines = content.splitlines()
        i = 0
        
        while i < len(lines):
            line = lines[i].strip()
            
            if not line or line.startswith('#') and not line.startswith('#EXTINF:'):
                i += 1
                continue
            
            title = None
            duration = None
            
            if line.startswith('#EXTINF:'):
                extinf_parts = line[8:].split(',', 1)
                try:
                    duration = int(float(extinf_parts[0]))
                except ValueError:
                    duration = None
                
                if len(extinf_parts) > 1:
                    title = extinf_parts[1].strip()
                
                i += 1
                if i >= len(lines):
                    break
                    
                line = lines[i].strip()
            
            if line and not line.startswith('#'):
                location = cls.normalize_path(line)
                entries.append(PlaylistEntry(location=location, title=title, duration=duration))
            
            i += 1
        
        return entries
    
    @classmethod
    def serialize(cls, entries: List[PlaylistEntry], original_content: Optional[str] = None) -> str:
        header = "#EXTM3U\n" if not original_content or original_content.strip().startswith("#EXTM3U") else ""
        lines = [header] if header else []
        
        for entry in entries:
            if entry.title or entry.duration is not None:
                duration_str = str(entry.duration) if entry.duration is not None else "-1"
                title = entry.title or ""
                lines.append(f"#EXTINF:{duration_str},{title}")
            
            lines.append(entry.location)
        
        return "\n".join(lines)


class M3U8Parser(M3UParser):
    @classmethod
    def serialize(cls, entries: List[PlaylistEntry], original_content: Optional[str] = None) -> str:
        content = super().serialize(entries, original_content)
        return content


class PLSParser(PlaylistParser):
    @classmethod
    def parse(cls, content: str) -> List[PlaylistEntry]:
        entries = []
        lines = content.splitlines()
        
        file_map = {}
        title_map = {}
        length_map = {}
        
        num_entries = 0
        
        for line in lines:
            line = line.strip()
            if not line:
                continue
                
            if line.lower().startswith('[playlist]'):
                continue
                
            if '=' not in line:
                continue
                
            key, value = line.split('=', 1)
            key = key.strip().lower()
            value = value.strip()
            
            if key.startswith('file'):
                try:
                    index = int(key[4:])
                    file_map[index] = cls.normalize_path(value)
                    num_entries = max(num_entries, index)
                except ValueError:
                    pass
                    
            elif key.startswith('title'):
                try:
                    index = int(key[5:])
                    title_map[index] = value
                except ValueError:
                    pass
                    
            elif key.startswith('length'):
                try:
                    index = int(key[6:])
                    try:
                        length_map[index] = int(float(value)) if value != '-1' else None
                    except ValueError:
                        length_map[index] = None
                except ValueError:
                    pass
        
        for i in range(1, num_entries + 1):
            if i in file_map:
                title = title_map.get(i)
                duration = length_map.get(i)
                entries.append(PlaylistEntry(location=file_map[i], title=title, duration=duration))
        
        return entries
    
    @classmethod
    def serialize(cls, entries: List[PlaylistEntry], original_content: Optional[str] = None) -> str:
        lines = ["[playlist]"]
        lines.append(f"NumberOfEntries={len(entries)}")
        
        for i, entry in enumerate(entries, 1):
            lines.append(f"File{i}={entry.location}")
            
            if entry.title:
                lines.append(f"Title{i}={entry.title}")
            else:
                lines.append(f"Title{i}=")
                
            duration = -1 if entry.duration is None else entry.duration
            lines.append(f"Length{i}={duration}")
        
        lines.append("Version=2")
        return "\n".join(lines)


class XSPFParser(PlaylistParser):
    @classmethod
    def parse(cls, content: str) -> List[PlaylistEntry]:
        try:
            import xml.etree.ElementTree as ET
            entries = []
            
            root = ET.fromstring(content)
            ns = {'ns': 'http://xspf.org/ns/0/'}
            
            for track in root.findall('.//ns:track', ns):
                location_elem = track.find('ns:location', ns)
                title_elem = track.find('ns:title', ns)
                creator_elem = track.find('ns:creator', ns)
                album_elem = track.find('ns:album', ns)
                duration_elem = track.find('ns:duration', ns)
                
                location = cls.normalize_path(location_elem.text.strip()) if location_elem is not None and location_elem.text else ""
                title = title_elem.text if title_elem is not None and title_elem.text else None
                artist = creator_elem.text if creator_elem is not None and creator_elem.text else None
                album = album_elem.text if album_elem is not None and album_elem.text else None
                
                duration = None
                if duration_elem is not None and duration_elem.text:
                    try:
                        duration = int(int(duration_elem.text) / 1000)  # Convert from milliseconds to seconds
                    except ValueError:
                        duration = None
                
                if location:
                    entries.append(PlaylistEntry(
                        location=location,
                        title=title,
                        artist=artist,
                        album=album,
                        duration=duration
                    ))
            
            return entries
        except Exception:
            return []
    
    @classmethod
    def serialize(cls, entries: List[PlaylistEntry], original_content: Optional[str] = None) -> str:
        try:
            import xml.etree.ElementTree as ET
            import xml.dom.minidom as minidom
            
            ET.register_namespace('', 'http://xspf.org/ns/0/')
            
            root = None
            if original_content:
                try:
                    root = ET.fromstring(original_content)
                except:
                    root = None
            
            if root is None:
                root = ET.Element('{http://xspf.org/ns/0/}playlist')
                ET.SubElement(root, '{http://xspf.org/ns/0/}title').text = "Playlist"
                ET.SubElement(root, '{http://xspf.org/ns/0/}creator').text = "Media Library"
                ET.SubElement(root, '{http://xspf.org/ns/0/}info')
                
            tracklist = root.find('.//{http://xspf.org/ns/0/}trackList')
            if tracklist is None:
                tracklist = ET.SubElement(root, '{http://xspf.org/ns/0/}trackList')
            else:
                for child in list(tracklist):
                    tracklist.remove(child)
            
            for entry in entries:
                track = ET.SubElement(tracklist, '{http://xspf.org/ns/0/}track')
                
                location = ET.SubElement(track, '{http://xspf.org/ns/0/}location')
                location.text = entry.location
                
                if entry.title:
                    title = ET.SubElement(track, '{http://xspf.org/ns/0/}title')
                    title.text = entry.title
                
                if entry.artist:
                    creator = ET.SubElement(track, '{http://xspf.org/ns/0/}creator')
                    creator.text = entry.artist
                
                if entry.album:
                    album = ET.SubElement(track, '{http://xspf.org/ns/0/}album')
                    album.text = entry.album
                
                if entry.duration is not None:
                    duration = ET.SubElement(track, '{http://xspf.org/ns/0/}duration')
                    duration.text = str(entry.duration * 1000)  # Convert to milliseconds
            
            xml_str = ET.tostring(root, encoding='unicode')
            dom = minidom.parseString(xml_str)
            formatted_xml = dom.toprettyxml(indent='  ')
            
            return formatted_xml
        except Exception as e:
            return f"<?xml version='1.0' encoding='UTF-8'?>\n<playlist xmlns='http://xspf.org/ns/0/'>\n  <title>Playlist</title>\n  <trackList>\n  </trackList>\n</playlist>"


class JSONParser(PlaylistParser):
    @classmethod
    def parse(cls, content: str) -> List[PlaylistEntry]:
        try:
            data = json.loads(content)
            entries = []
            
            if isinstance(data, dict) and "tracks" in data and isinstance(data["tracks"], list):
                for track in data["tracks"]:
                    if isinstance(track, dict) and "location" in track:
                        entry = PlaylistEntry(
                            location=cls.normalize_path(track["location"]),
                            title=track.get("title"),
                            artist=track.get("artist"),
                            album=track.get("album"),
                            duration=track.get("duration"),
                            metadata={k: v for k, v in track.items() if k not in ["location", "title", "artist", "album", "duration"]}
                        )
                        entries.append(entry)
            elif isinstance(data, list):
                for track in data:
                    if isinstance(track, dict) and "location" in track:
                        entry = PlaylistEntry(
                            location=cls.normalize_path(track["location"]),
                            title=track.get("title"),
                            artist=track.get("artist"),
                            album=track.get("album"),
                            duration=track.get("duration"),
                            metadata={k: v for k, v in track.items() if k not in ["location", "title", "artist", "album", "duration"]}
                        )
                        entries.append(entry)
            
            return entries
        except:
            return []
    
    @classmethod
    def serialize(cls, entries: List[PlaylistEntry], original_content: Optional[str] = None) -> str:
        structure_type = "object"  # Default
        
        if original_content:
            try:
                data = json.loads(original_content)
                if isinstance(data, list):
                    structure_type = "array"
            except:
                pass
        
        if structure_type == "array":
            tracks = []
            for entry in entries:
                track = {
                    "location": entry.location
                }
                
                if entry.title:
                    track["title"] = entry.title
                
                if entry.artist:
                    track["artist"] = entry.artist
                
                if entry.album:
                    track["album"] = entry.album
                
                if entry.duration is not None:
                    track["duration"] = entry.duration
                
                if entry.metadata:
                    track.update(entry.metadata)
                
                tracks.append(track)
            
            return json.dumps(tracks, indent=2)
        else:
            playlist = {"tracks": []}
            
            for entry in entries:
                track = {
                    "location": entry.location
                }
                
                if entry.title:
                    track["title"] = entry.title
                
                if entry.artist:
                    track["artist"] = entry.artist
                
                if entry.album:
                    track["album"] = entry.album
                
                if entry.duration is not None:
                    track["duration"] = entry.duration
                
                if entry.metadata:
                    track.update(entry.metadata)
                
                playlist["tracks"].append(track)
            
            return json.dumps(playlist, indent=2)


class Playlist:
    def __init__(self, title: Optional[str] = None):
        self.title = title or "New Playlist"
        self.entries: List[PlaylistEntry] = []
        self._original_content: Optional[str] = None
        self._format: Optional[PlaylistFormat] = None
        self._parsers: Dict[PlaylistFormat, Type[PlaylistParser]] = {
            PlaylistFormat.M3U: M3UParser,
            PlaylistFormat.M3U8: M3U8Parser,
            PlaylistFormat.XSPF: XSPFParser,
            PlaylistFormat.PLS: PLSParser,
            PlaylistFormat.JSON: JSONParser
        }
        self._custom_parsers: Dict[str, Type[PlaylistParser]] = {}
    
    def register_parser(self, format_type: Union[PlaylistFormat, str], parser_class: Type[PlaylistParser]) -> None:
        if isinstance(format_type, PlaylistFormat):
            self._parsers[format_type] = parser_class
        else:
            self._custom_parsers[format_type] = parser_class
    
    def _detect_format(self, file_path: Union[str, Path]) -> PlaylistFormat:
        file_path = str(file_path)
        file_path_lower = file_path.lower()
        
        if file_path_lower.endswith('.m3u8'):
            return PlaylistFormat.M3U8
        elif file_path_lower.endswith('.m3u'):
            return PlaylistFormat.M3U
        elif file_path_lower.endswith('.xspf'):
            return PlaylistFormat.XSPF
        elif file_path_lower.endswith('.pls'):
            return PlaylistFormat.PLS
        elif file_path_lower.endswith('.json'):
            return PlaylistFormat.JSON
        else:
            return PlaylistFormat.UNKNOWN
    
    def load(self, file_path_or_url: Union[str, Path], format_type: Optional[PlaylistFormat] = None, encoding: str = 'utf-8') -> 'Playlist':
        file_path_str = str(file_path_or_url)
        
        if not format_type:
            format_type = self._detect_format(file_path_str)
            
            if format_type == PlaylistFormat.UNKNOWN:
                raise ValueError(f"Cannot determine playlist format for file: {file_path_str}")
        
        self._format = format_type
        
        if re.match(r'^https?://', file_path_str):
            with urllib.request.urlopen(file_path_str) as response:
                content = response.read().decode(encoding)
        else:
            with open(file_path_str, 'r', encoding=encoding) as f:
                content = f.read()
        
        self._original_content = content
        
        parser_class = self._parsers.get(format_type)
        if not parser_class:
            raise ValueError(f"No parser registered for format: {format_type}")
        
        self.entries = parser_class.parse(content)
        return self
    
    def save(self, file_path: Union[str, Path], format_type: Optional[PlaylistFormat] = None, encoding: str = 'utf-8') -> None:
        file_path_str = str(file_path)
        
        if not format_type:
            if self._format:
                format_type = self._format
            else:
                format_type = self._detect_format(file_path_str)
                
                if format_type == PlaylistFormat.UNKNOWN:
                    raise ValueError(f"Cannot determine playlist format for file: {file_path_str}")
        
        parser_class = self._parsers.get(format_type)
        if not parser_class:
            raise ValueError(f"No parser registered for format: {format_type}")
        
        content = parser_class.serialize(self.entries, self._original_content if format_type == self._format else None)
        
        with open(file_path_str, 'w', encoding=encoding) as f:
            f.write(content)
    
    def add_entry(self, entry: PlaylistEntry) -> None:
        self.entries.append(entry)
    
    def remove_entry(self, index: int) -> Optional[PlaylistEntry]:
        if 0 <= index < len(self.entries):
            return self.entries.pop(index)
        return None
    
    def move_entry(self, from_index: int, to_index: int) -> bool:
        if 0 <= from_index < len(self.entries) and 0 <= to_index < len(self.entries):
            entry = self.entries.pop(from_index)
            self.entries.insert(to_index, entry)
            return True
        return False
    
    def clear(self) -> None:
        self.entries.clear()
    
    def get_entry(self, index: int) -> Optional[PlaylistEntry]:
        if 0 <= index < len(self.entries):
            return self.entries[index]
        return None
    
    def sort(self, key: Union[str, Callable[[PlaylistEntry], Any]], reverse: bool = False) -> None:
        if isinstance(key, str):
            attr = key
            key = lambda entry: getattr(entry, attr) if getattr(entry, attr) is not None else ""
        
        self.entries.sort(key=key, reverse=reverse)
    
    def filter(self, predicate: Callable[[PlaylistEntry], bool]) -> 'Playlist':
        result = Playlist(self.title)
        result.entries = [entry for entry in self.entries if predicate(entry)]
        result._format = self._format
        result._original_content = self._original_content
        return result
    
    def remove_duplicates(self, key: Union[str, Callable[[PlaylistEntry], Any]] = "location") -> int:
        if isinstance(key, str):
            attr = key
            key = lambda entry: getattr(entry, attr)
        
        seen = set()
        unique_entries = []
        removed_count = 0
        
        for entry in self.entries:
            entry_key = key(entry)
            if entry_key not in seen:
                seen.add(entry_key)
                unique_entries.append(entry)
            else:
                removed_count += 1
        
        self.entries = unique_entries
        return removed_count
    
    def filter_by_extension(self, allowed_extensions: List[str], include: bool = True) -> int:
        allowed_extensions = [ext.lower() for ext in allowed_extensions]
        filtered_count = 0
        
        def has_allowed_extension(entry: PlaylistEntry) -> bool:
            location = entry.location.lower()
            file_ext = os.path.splitext(location)[1][1:] if os.path.splitext(location)[1] else ""
            return (file_ext in allowed_extensions) == include
        
        original_count = len(self.entries)
        self.entries = [entry for entry in self.entries if has_allowed_extension(entry)]
        filtered_count = original_count - len(self.entries)
        
        return filtered_count
    
    def merge(self, other: 'Playlist') -> 'Playlist':
        merged = Playlist(f"{self.title} + {other.title}")
        merged.entries = self.entries.copy()
        merged.entries.extend(other.entries)
        return merged
    
    def __len__(self) -> int:
        return len(self.entries)
    
    def __getitem__(self, index: int) -> PlaylistEntry:
        return self.entries[index]
    
    def __iter__(self):
        return iter(self.entries)


class PlaylistManager:
    def __init__(self):
        self.playlists: Dict[str, Playlist] = {}
    
    def create_playlist(self, name: str, title: Optional[str] = None) -> Playlist:
        playlist = Playlist(title)
        self.playlists[name] = playlist
        return playlist
    
    def get_playlist(self, name: str) -> Optional[Playlist]:
        return self.playlists.get(name)
    
    def remove_playlist(self, name: str) -> bool:
        if name in self.playlists:
            del self.playlists[name]
            return True
        return False
    
    def load_playlist(self, name: str, file_path_or_url: Union[str, Path], 
                     format_type: Optional[PlaylistFormat] = None,
                     encoding: str = 'utf-8') -> Playlist:
        playlist = Playlist()
        playlist.load(file_path_or_url, format_type, encoding)
        self.playlists[name] = playlist
        return playlist
    
    def save_playlist(self, name: str, file_path: Union[str, Path], 
                     format_type: Optional[PlaylistFormat] = None,
                     encoding: str = 'utf-8') -> bool:
        playlist = self.get_playlist(name)
        if playlist:
            playlist.save(file_path, format_type, encoding)
            return True
        return False
    
    def list_playlists(self) -> List[str]:
        return list(self.playlists.keys())
    
    def merge_playlists(self, name: str, source_playlist_names: List[str], 
                       title: Optional[str] = None) -> Optional[Playlist]:
        if not source_playlist_names:
            return None
        
        source_playlists = [self.get_playlist(name) for name in source_playlist_names]
        
        if any(playlist is None for playlist in source_playlists):
            return None
        
        merged_playlist = source_playlists[0]
        for playlist in source_playlists[1:]:
            merged_playlist = merged_playlist.merge(playlist)
        
        if title:
            merged_playlist.title = title
            
        self.playlists[name] = merged_playlist
        return merged_playlist
    
    def sort_playlist(self, name: str, key: Union[str, Callable[[PlaylistEntry], Any]], 
                     reverse: bool = False) -> bool:
        playlist = self.get_playlist(name)
        if playlist:
            playlist.sort(key=key, reverse=reverse)
            return True
        return False
    
    def filter_playlist(self, source_name: str, target_name: str, 
                       predicate: Callable[[PlaylistEntry], bool],
                       title: Optional[str] = None) -> Optional[Playlist]:
        source_playlist = self.get_playlist(source_name)
        if not source_playlist:
            return None
        
        filtered_playlist = source_playlist.filter(predicate)
        if title:
            filtered_playlist.title = title
            
        self.playlists[target_name] = filtered_playlist
        return filtered_playlist
    
    def remove_duplicates(self, name: str, key: Union[str, Callable[[PlaylistEntry], Any]] = "location") -> int:
        playlist = self.get_playlist(name)
        if playlist:
            return playlist.remove_duplicates(key)
        return 0
    
    def filter_by_extension(self, name: str, allowed_extensions: List[str], include: bool = True) -> int:
        playlist = self.get_playlist(name)
        if playlist:
            return playlist.filter_by_extension(allowed_extensions, include)
        return 0



from setuptools import setup, find_packages


def get_readme():
    with open("./readme.md", 'r') as file:
        return file.read()


setup(
    name="AVPlay",
    version="1.2.0",
    description="Audio and video playback library",
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/still-standing88/av-play/',
    license='MIT',
    author="still-standing88",
    packages=find_packages(),
    entry_points={
    },
    install_requires=["python-mpv", "python-vlc", "pyfmodex", "validators", "music-tag"],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Documentation",
        "Topic :: Utilities",
    ],
    keywords="av-play AVPlay audio video media playback",
)



from abc import ABC, abstractmethod
from .__AV_Common import *
from .__AV_Instance import AVMediaInstance
from .__AV_Interface import AVMediaInterface
from .playlist import Playlist
import threading
import time
import random
from typing import List, Optional, Callable


class AVPlayer(ABC):


    def __init__(self, media_type:AVMediaType, media_backend:AVMediaBackend, interface:AVMediaInterface) -> None:
        super().__init__()
        self._media_type = media_type
        self._media_backend = media_backend
        self._controler = interface
        self._config:dict[str, Any] = {}
        self._primary_instance:AVMediaInstance|None = None
        self._current_playlist:Playlist|None = None
        self._current_playlist_index = -1
        

        self._auto_play_enabled = False
        self._playlist_mode = AVPlaylistMode.SEQUENTIAL
        self._playlist_state = AVPlaylistState.STOPPED
        self._monitor_thread:threading.Thread|None = None
        self._monitor_running = False
        self._shuffle_order:List[int] = []
        self._track_end_callback:Optional[Callable[[int], None]] = None


    @property
    def media_type(self):
        return self._media_type

    @property
    def media_backend(self):
        return self._media_backend

    @property
    def config(self):
        return self._config

    @property
    def primary_instance(self):
            return self._primary_instance

    @property
    def current_playlist(self):
        return self._current_playlist


    def create_file_instance(self, file_path:str) -> AVMediaInstance:
        instance:AVMediaInstance = AVMediaInstance(self._controler)
        instance.load_file(file_path)
        return instance

    def create_url_instance(self, url:str) -> AVMediaInstance:
        instance:AVMediaInstance = AVMediaInstance(self._controler)
        instance.load_url(url)
        return instance

    def set_config_value(self, name:str, value:Any):
        self._config[name] = value

    def get_config(self, name) -> Any:
        if name in self._config:
            return self._config[name]
        return None

    def load_playlist(self, playlist:Playlist, auto_play:bool = False, mode:AVPlaylistMode = AVPlaylistMode.SEQUENTIAL):
        """Load playlist with optional auto-play functionality"""
        if playlist is not None:
            self._current_playlist = playlist
            self._auto_play_enabled = auto_play
            self._playlist_mode = mode
            self._playlist_state = AVPlaylistState.STOPPED


            if mode == AVPlaylistMode.SHUFFLE:
                self._generate_shuffle_order()

            if self._primary_instance is not None:
                self._primary_instance.stop()
            else:
                self._primary_instance = AVMediaInstance(self._controler)
                
            if len(playlist) > 0:
                self._current_playlist_index = 0
                self._play_playlist_track()
                

                if auto_play:
                    self._start_monitor()

    def set_playlist_mode(self, mode:AVPlaylistMode):
        """Change playlist playback mode"""
        self._playlist_mode = mode
        if mode == AVPlaylistMode.SHUFFLE:
            self._generate_shuffle_order()

    def set_auto_play(self, enabled:bool):
        """Enable/disable auto-play functionality"""
        self._auto_play_enabled = enabled
        if enabled and self._current_playlist and len(self._current_playlist) > 0:
            self._start_monitor()
        else:
            self._stop_monitor()

    def set_track_end_callback(self, callback:Optional[Callable[[int], None]]):
        """Set callback function called when a track ends"""
        self._track_end_callback = callback

    def get_playlist_state(self) -> AVPlaylistState:
        """Get current playlist state"""
        return self._playlist_state

    def get_playlist_mode(self) -> AVPlaylistMode:
        """Get current playlist mode"""
        return self._playlist_mode

    def is_auto_play_enabled(self) -> bool:
        """Check if auto-play is enabled"""
        return self._auto_play_enabled

    def get_current_track_index(self) -> int:
        """Get current track index in playlist"""
        return self._current_playlist_index

    def previous(self):
        if self._current_playlist is not None and len(self._current_playlist) > 0:
            if self._playlist_mode == AVPlaylistMode.SHUFFLE:

                current_shuffle_pos = self._shuffle_order.index(self._current_playlist_index)
                current_shuffle_pos = max(0, current_shuffle_pos - 1)
                self._current_playlist_index = self._shuffle_order[current_shuffle_pos]
            else:
                self._current_playlist_index = max(0, self._current_playlist_index - 1)
            self._play_playlist_track()

    def next(self):
        if self._current_playlist is not None and len(self._current_playlist) > 0:
            self._advance_track()

    def stop_playlist(self):
        """Stop playlist playback"""
        if self._primary_instance:
            self._primary_instance.stop()
        self._playlist_state = AVPlaylistState.STOPPED
        self._stop_monitor()

    def pause_playlist(self):
        """Pause playlist playback"""
        if self._primary_instance:
            self._primary_instance.pause()
        self._playlist_state = AVPlaylistState.PAUSED

    def resume_playlist(self):
        """Resume playlist playback"""
        if self._primary_instance:
            self._primary_instance.play()
        self._playlist_state = AVPlaylistState.PLAYING
        if self._auto_play_enabled:
            self._start_monitor()

    def _generate_shuffle_order(self):
        """Generate random order for shuffle mode"""
        if self._current_playlist:
            self._shuffle_order = list(range(len(self._current_playlist)))
            random.shuffle(self._shuffle_order)

    def _advance_track(self):
        """Advance to next track based on current mode"""
        if not self._current_playlist or len(self._current_playlist) == 0:
            return

        if self._playlist_mode == AVPlaylistMode.REPEAT_ONE:

            pass
        elif self._playlist_mode == AVPlaylistMode.SHUFFLE:

            try:
                current_shuffle_pos = self._shuffle_order.index(self._current_playlist_index)
                current_shuffle_pos += 1
                if current_shuffle_pos >= len(self._shuffle_order):
                    if self._playlist_mode == AVPlaylistMode.REPEAT_ALL:
                        current_shuffle_pos = 0
                    else:
                        self._playlist_state = AVPlaylistState.FINISHED
                        return
                self._current_playlist_index = self._shuffle_order[current_shuffle_pos]
            except ValueError:
                self._current_playlist_index = self._shuffle_order[0] if self._shuffle_order else 0
        else:

            self._current_playlist_index += 1
            if self._current_playlist_index >= len(self._current_playlist):
                if self._playlist_mode == AVPlaylistMode.REPEAT_ALL:
                    self._current_playlist_index = 0
                else:
                    self._playlist_state = AVPlaylistState.FINISHED
                    return

        self._play_playlist_track()

    def _start_monitor(self):
        """Start background thread to monitor track completion"""
        if not self._monitor_running:
            self._monitor_running = True
            self._monitor_thread = threading.Thread(target=self._monitor_playback, daemon=True)
            self._monitor_thread.start()

    def _stop_monitor(self):
        """Stop background monitoring thread"""
        self._monitor_running = False
        if self._monitor_thread and self._monitor_thread.is_alive():
            self._monitor_thread.join(timeout=1.0)

    def _monitor_playback(self):
        """Background thread that monitors track completion"""
        while self._monitor_running and self._auto_play_enabled:
            try:
                if self._primary_instance:
                    state = self._primary_instance.get_playback_state()
                    
                    if state == AVPlaybackState.AV_STATE_PLAYING:
                        self._playlist_state = AVPlaylistState.PLAYING
                        

                        position = self._primary_instance.get_position()
                        length = self._primary_instance.get_length()
                        
                        if length > 0 and position >= length - 1:

                            current_track = self._current_playlist_index
                            if self._track_end_callback:
                                self._track_end_callback(current_track)
                            
                            time.sleep(0.5)
                            self._advance_track()
                            
                    elif state in [AVPlaybackState.AV_STATE_STOPPED, AVPlaybackState.AV_STATE_NOTHING]:
                        if self._playlist_state == AVPlaylistState.PLAYING:

                            self._advance_track()
                
                time.sleep(0.1)
                
            except Exception:

                time.sleep(0.5)

    def _play_playlist_track(self):
        if self._primary_instance is None:
            self._primary_instance = AVMediaInstance(self._controler)
        
        if self._current_playlist and 0 <= self._current_playlist_index < len(self._current_playlist):
            instance_path = self._current_playlist.entries[self._current_playlist_index].location
            if is_path(instance_path):
                self._primary_instance.load_file(instance_path)
            else:
                self._primary_instance.load_url(instance_path)
            self._primary_instance.play()
            self._playlist_state = AVPlaylistState.PLAYING


    @abstractmethod
    def init(self, *args, **kw):
        pass

    @abstractmethod
    def release(self):

        self._stop_monitor()
        pass

    def set_device(self, index:int):
        device_count:int  = self._controler.get_devices()
        if index < device_count and index >= 0:
            self._controler.set_device(index)
        else:
            raise IndexError

    def get_device(self, index:int) -> AVDevice | None:
        device_count:int  = self._controler.get_devices()
        if index < device_count and index >= 0:
            return self._controler.get_device_info(index)
        else:
            raise IndexError

    def get_devices(self) -> int:
        return self._controler.get_devices()

    def get_current_device(self) -> int:
        return self._controler.get_current_device()





import mpv
import os
from typing import Dict, Any, Union, List, Callable
from .__AV_Common import *
from .__AV_Instance import AVMediaInstance
from .__AV_Interface import AVMediaInterface
from .__AV_Player import AVPlayer
from .mpv_audio_filter import MPVAudioFilter

def handle_mpv_error(call_func: Callable) -> Any:
    try:
        return call_func()
    except mpv.PropertyUnavailableError as e:
        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER, f"Property unavailable: {str(e)}")
    except AttributeError as e:
        if "mpv property does not exist" in str(e) or "does not exist" in str(e):
            raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER, f"Property not available: {str(e)}")
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"Attribute error: {str(e)}")
    except mpv.ShutdownError:
        raise AVError(AVErrorInfo.INVALID_HANDLE, "MPV core has been shutdown")
    except RuntimeError as e:
        if "loading failed" in str(e).lower():
            raise AVError(AVErrorInfo.FILE_NOTFOUND, f"File loading failed: {str(e)}")
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"Runtime error: {str(e)}")
    except SystemError as e:
        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER, f"Command error: {str(e)}")
    except TypeError as e:
        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_TYPE, f"Type error: {str(e)}")
    except ValueError as e:
        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_VALUE, f"Value error: {str(e)}")
    except Exception as e:
        raise AVError(AVErrorInfo.UNKNOWN_ERROR, f"Unknown error: {str(e)}")

class MPVMediaInterface(AVMediaInterface):
    
    def __init__(self) -> None:
        super().__init__(AVMediaType.AV_TYPE_VIDEO, AVMediaBackend.AV_BACKEND_MPV)
        self.__instances:dict[int, str] = {}
        self.__mpv: mpv.MPV | None = None
        self.__current_id: int | None = None
        self.__applied_filters: Dict[int, MPVAudioFilter] = {}
        self.__is_initialized = False
        self.__stopped = False
        self.__end_reached = False

    def init(self, *args, **kw):
        try:
            config = kw.get("config", {})
            window = kw.get("window", None)
            
            if window is not None:
                config['wid'] = str(int(window))
            
            config.setdefault('ytdl', True)
            config.setdefault('input_default_bindings', True)
            
            self.__mpv = mpv.MPV(**config)
            
            ytdl_path = kw.get("ytdl_path", None)
            if ytdl_path:
                def set_ytdl_path():
                    assert self.__mpv is not None
                    self.__mpv.script_opts["ytdl_hook-ytdl_path"] = ytdl_path
                handle_mpv_error(set_ytdl_path)
            
            self.__is_initialized = True
            self.__stopped = False
            self.__end_reached = False
            
        except Exception as e:
            raise AVError(AVErrorInfo.INITIALIZATION_ERROR, f"Failed to initialize MPV: {str(e)}")

    def free(self):
        if self.__mpv is not None:
            try:
                if not self.__mpv.core_shutdown:
                    self.__mpv.terminate()
                self.__mpv = None
                self.__current_id = None
                self.__applied_filters.clear()
                self.__is_initialized = False
                self.__stopped = False
                self.__end_reached = False
            except Exception:
                pass

    def _check_initialized(self):
        if not self.__is_initialized or self.__mpv is None:
            raise AVError(AVErrorInfo.UNINITIALIZED, "MPV not initialized")
        if self.__mpv.core_shutdown:
            raise AVError(AVErrorInfo.INVALID_HANDLE, "MPV core has been shutdown")

    def _check_instance(self, id: int):
        self._check_initialized()
        if self.__current_id != id:
            raise AVError(AVErrorInfo.INVALID_HANDLE, f"Invalid instance ID: {id}")

    def get_mpv_instance(self):
        self._check_initialized()
        return self.__mpv

    def load_file(self, id: int, path: str):
        self._check_initialized()

        def loadfile():
            assert self.__mpv is not None
            self.__mpv.command("loadfile", path)
        
        if path and is_path(path) and os.path.exists(path):
            self.__current_id = id
            self.__instances[id] = path
            self.__stopped = False
            self.__end_reached = False
            self._load_subtitles(path)
            handle_mpv_error(loadfile)

    def load_url(self, id: int, url: str):
        self._check_initialized()
        
        def loadurl():
            assert self.__mpv is not None
            self.__mpv.command("loadfile", url)
        
        if url and is_url(url):
            self.__current_id = id
            self.__instances[id] = url
            self.__stopped = False
            self.__end_reached = False
            handle_mpv_error(loadurl)

    def _load_subtitles(self, path: str):
        subtitle_formats = ["idx", "sub", "srt", "rt", "ssa", "ass", "mks", "vtt", "sup", "scc", "smi", "lrc", "pgs"]
        dir_path = os.path.dirname(path)
        filename = os.path.splitext(os.path.basename(path))[0]
        
        for format in subtitle_formats:
            subtitle_path = os.path.join(dir_path, f"{filename}.{format}")
            if os.path.exists(subtitle_path):
                def add_subtitle():
                    assert self.__mpv is not None
                    self.__mpv.sub_add(subtitle_path)
                handle_mpv_error(add_subtitle)
                break

    def release(self, id: int):
        if self.__current_id == id:
            self._check_initialized()
            
            def stop():
                assert self.__mpv is not None
                self.__mpv.stop()
            handle_mpv_error(stop)
            self.__current_id = None
            self.__applied_filters.clear()
            self.__stopped = True
            self.__end_reached = False

    def play(self, id: int):
        self._check_instance(id)
        
        def resume():
            assert self.__mpv is not None
            if self.__end_reached or self.__mpv.time_pos is None:
                self.__mpv.command("loadfile", self.__instances[id])
                self.__end_reached = False
                self.__stopped = False
            self.__mpv.pause = False
        
        self.__stopped = False
        handle_mpv_error(resume)

    def pause(self, id: int):
        self._check_instance(id)
        
        def pause_play():
            assert self.__mpv is not None
            self.__mpv.pause = True
        handle_mpv_error(pause_play)

    def mute(self, id: int):
        self._check_instance(id)
        
        def toggle_mute():
            assert self.__mpv is not None
            self.__mpv.mute = not self.__mpv.mute
        handle_mpv_error(toggle_mute)

    def unmute(self, id: int):
        self._check_instance(id)
        
        def unmute_audio():
            assert self.__mpv is not None
            self.__mpv.mute = False
        handle_mpv_error(unmute_audio)

    def stop(self, id: int):
        self._check_instance(id)
        
        def halt():
            assert self.__mpv is not None
            self.__mpv.stop()
        
        handle_mpv_error(halt)
        self.__stopped = True
        self.__end_reached = False

    def set_volume(self, id: int, offset: float):
        self._check_instance(id)
        
        def set_vol():
            assert self.__mpv is not None
            self.__mpv.volume = offset
        handle_mpv_error(set_vol)

    def set_position(self, id: int, offset: int):
        self._check_instance(id)
        
        def seek():
            assert self.__mpv is not None
            self.__mpv.seek(offset, "absolute")
        handle_mpv_error(seek)

    def set_loop(self, id: int, loop: bool):
        self._check_instance(id)
        loop_value = "inf" if loop else "no"
        
        def set_loop_mode():
            assert self.__mpv is not None
            self.__mpv.loop = loop_value
        handle_mpv_error(set_loop_mode)

    def get_length(self, id: int) -> int:
        self._check_instance(id)
        
        def get_duration():
            assert self.__mpv is not None
            return self.__mpv.duration
        duration = handle_mpv_error(get_duration)
        return int(duration) if duration is not None else 0

    def get_position(self, id: int) -> int:
        self._check_instance(id)
        
        def get_time():
            assert self.__mpv is not None
            return self.__mpv.time_pos
        pos = handle_mpv_error(get_time)
        return int(pos) if pos is not None else 0

    def get_play_state(self, id: int) -> AVPlaybackState:
        self._check_instance(id)
        try:
            def get_pause():
                assert self.__mpv is not None
                return self.__mpv.pause
            def get_time_pos():
                assert self.__mpv is not None
                return self.__mpv.time_pos
            def get_duration():
                assert self.__mpv is not None
                return self.__mpv.duration
            
            if self.__stopped:
                return AVPlaybackState.AV_STATE_STOPPED
            
            is_paused = handle_mpv_error(get_pause)
            time_pos = handle_mpv_error(get_time_pos)
            duration = handle_mpv_error(get_duration)
            
            if duration is not None and time_pos is not None:
                if time_pos >= duration - 0.1:
                    self.__end_reached = True
                    return AVPlaybackState.AV_STATE_NOTHING
            elif time_pos is None and duration is not None:
                self.__end_reached = True
                return AVPlaybackState.AV_STATE_NOTHING
            
            if is_paused:
                return AVPlaybackState.AV_STATE_PAUSED
            else:
                return AVPlaybackState.AV_STATE_PLAYING
                
        except AVError:
            return AVPlaybackState.AV_STATE_NOTHING

    def get_mute_state(self, id: int) -> AVMuteState:
        self._check_instance(id)
        
        def get_mute():
            assert self.__mpv is not None
            return self.__mpv.mute
        muted = handle_mpv_error(get_mute)
        return AVMuteState.AV_AUDIO_MUTED if muted else AVMuteState.AV_AUDIO_UNMUTED

    def get_volume(self, id: int) -> float:
        self._check_instance(id)
        
        def get_vol():
            assert self.__mpv is not None
            return self.__mpv.volume
        volume = handle_mpv_error(get_vol)
        return float(volume) if volume is not None else 0.0

    def get_loop(self, id: int) -> bool:
        self._check_instance(id)
        
        def get_loop_state():
            assert self.__mpv is not None
            return self.__mpv.loop
        loop_val = handle_mpv_error(get_loop_state)
        return loop_val == "inf" if loop_val is not None else False

    def _rebuild_filter_chain(self):
        filter_strings = []
        
        for filter_obj in self.__applied_filters.values():
            filter_string = filter_obj.construct()
            if filter_string:
                filter_strings.append(filter_string)
        
        return ",".join(filter_strings) if filter_strings else ""

    def apply_filter(self, id: int, filter_id: int, filter_struct: AVFilter):
        self._check_instance(id)
        if not isinstance(filter_struct, MPVAudioFilter):
            raise AVError(AVErrorInfo.INVALID_MEDIA_FILTER, "Filter must be MPVAudioFilter")
        
        self.__applied_filters[filter_id] = filter_struct
        
        filter_chain = self._rebuild_filter_chain()
        
        def set_af_chain():
            assert self.__mpv is not None
            self.__mpv.af = filter_chain
        
        handle_mpv_error(set_af_chain)

    def remove_filter(self, id: int, filter_id: int):
        self._check_instance(id)
        if filter_id in self.__applied_filters:
            del self.__applied_filters[filter_id]
            
            filter_chain = self._rebuild_filter_chain()
            
            def set_af_chain():
                assert self.__mpv is not None
                self.__mpv.af = filter_chain
            
            handle_mpv_error(set_af_chain)

    def set_parameter(self, id: int, filter_id: int, parameter_name: str, value: ParameterValue):
        self._check_instance(id)
        if filter_id in self.__applied_filters:
            filter_struct = self.__applied_filters[filter_id]
            filter_struct.set_parameter(parameter_name, value)
            
            filter_chain = self._rebuild_filter_chain()
            
            def set_af_chain():
                assert self.__mpv is not None
                self.__mpv.af = filter_chain
            
            handle_mpv_error(set_af_chain)

    def get_parameter(self, id: int, filter_id: int, parameter_name: str) -> ParameterValue | None:
        self._check_instance(id)
        if filter_id in self.__applied_filters:
            return self.__applied_filters[filter_id].get_parameter(parameter_name)
        return None

    def get_devices(self) -> int:
        self._check_initialized()
        
        def get_device_list():
            assert self.__mpv is not None
            return self.__mpv.audio_device_list
        audio_devices = handle_mpv_error(get_device_list) or []
        return len(audio_devices)

    def get_device_info(self, index: int) -> AVDevice:
        self._check_initialized()
        
        def get_device_list():
            assert self.__mpv is not None
            return self.__mpv.audio_device_list
        audio_devices = handle_mpv_error(get_device_list) or []
        if 0 <= index < len(audio_devices):
            device = audio_devices[index]
            return AVDevice(AVMediaBackend.AV_BACKEND_MPV, device.get('description', device.get('name', 'Unknown')))
        raise IndexError("Device index out of range")

    def set_device(self, index: int):
        self._check_initialized()
        
        def get_device_list():
            assert self.__mpv is not None
            return self.__mpv.audio_device_list
        def set_audio_device(device_name):
            assert self.__mpv is not None
            self.__mpv.audio_device = device_name
        audio_devices = handle_mpv_error(get_device_list) or []
        if 0 <= index < len(audio_devices):
            device_name = audio_devices[index]['name']
            handle_mpv_error(lambda: set_audio_device(device_name))
        else:
            raise IndexError("Device index out of range")

    def get_current_device(self) -> int:
        self._check_initialized()
        
        def get_audio_device():
            assert self.__mpv is not None
            return self.__mpv.audio_device
        def get_device_list():
            assert self.__mpv is not None
            return self.__mpv.audio_device_list
        current_device = handle_mpv_error(get_audio_device)
        audio_devices = handle_mpv_error(get_device_list) or []
        
        for i, device in enumerate(audio_devices):
            if device['name'] == current_device:
                return i
        return 0


class MPVVideoPlayer(AVPlayer):
    
    def __init__(self) -> None:
        self.__mpv_interface:MPVMediaInterface = MPVMediaInterface()
        super().__init__(AVMediaType.AV_TYPE_VIDEO, AVMediaBackend.AV_BACKEND_MPV, self.__mpv_interface)

    def init(self, *args, **kw):
        self._controler.init(*args, **kw)

    def release(self):
        if self._primary_instance is not None:
            self._primary_instance.release()
            self._primary_instance = None
        self._controler.free()

    def create_file_instance(self, file_path:str) -> AVMediaInstance:
        if self._primary_instance is None:
            self._primary_instance = AVMediaInstance(self._controler)
        self._primary_instance.load_file(file_path)
        return self._primary_instance

    def create_url_instance(self, url:str) -> AVMediaInstance:
        if self._primary_instance is None:
            self._primary_instance = AVMediaInstance(self._controler)
        self._primary_instance.load_url(url)
        return self._primary_instance

    def set_window(self, window):
        def set_wid():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                mpv_instance._set_property("wid", str(int(window)))
        handle_mpv_error(set_wid)

    def forward(self, offset):
        def seek_forward():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                mpv_instance.seek(+offset, reference='relative')
        handle_mpv_error(seek_forward)

    def backward(self, offset):
        def seek_backward():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                mpv_instance.seek(-offset, reference='relative')
        handle_mpv_error(seek_backward)

    def set_volume_relative(self, direction, offset):
        def adjust_volume():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                current_volume = mpv_instance.volume
                if direction == "up":
                    mpv_instance.volume = current_volume + offset
                elif direction == "down":
                    mpv_instance.volume = current_volume - offset
        handle_mpv_error(adjust_volume)


    def set_fullscreen(self, state):
        def toggle_fullscreen():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                mpv_instance.fullscreen = state
        handle_mpv_error(toggle_fullscreen)

    def get_fullscreen(self):
        def get_fs_state():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                return mpv_instance.fullscreen
            return False
        return handle_mpv_error(get_fs_state)

    def set_playback_speed(self, speed):
        def set_speed():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                mpv_instance.speed = speed
        handle_mpv_error(set_speed)

    def set_resolution(self, width, height):
        def set_res():
            mpv_instance = self.__mpv_interface.get_mpv_instance()
            if mpv_instance:
                mpv_instance.vf = f"scale={width}:{height}"
        handle_mpv_error(set_res)


from typing import Any
from .__AV_Common import *
from .audio_filter import AudioFilter

class VLCAudioFilter(AVFilter):
    def __init__(self, filter_handle: str, parameters: dict[str, ParameterValue], backend_additional_info: dict[Any, Any]):
        super().__init__(AVFilterType.AV_TYPE_AUDIO, AVMediaBackend.AV_BACKEND_VLC, filter_handle, parameters, backend_additional_info)
        self._validate_parameters()
    
    def _validate_parameters(self):
        vlc_param_map = self.info.get("vlc_param_map", {})
        current_params = self.get_parameters()
        
        for param_name, param_value in current_params.items():
            if param_name in vlc_param_map:
                vlc_param_name, param_type, param_range = vlc_param_map[param_name]
                
                if not isinstance(param_value, param_type):
                    raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_TYPE, 
                                f"Parameter {param_name} must be of type {param_type.__name__}")
                
                if param_range and param_type in [int, float]:
                    min_val, max_val = param_range[:2]
                    if not (min_val <= param_value <= max_val):
                        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_RANGE,
                                    f"Parameter {param_name} must be between {min_val} and {max_val}")
    
    def set_parameter(self, name: str, value: ParameterValue):
        vlc_param_map = self.info.get("vlc_param_map", {})
        if name in vlc_param_map:
            vlc_param_name, param_type, param_range = vlc_param_map[name]
            
            if not isinstance(value, param_type):
                raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_TYPE,
                            f"Parameter {name} must be of type {param_type.__name__}")
            
            if param_range and param_type in [int, float]:
                min_val, max_val = param_range[:2]
                if not (min_val <= value <= max_val):
                    raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_RANGE,
                                f"Parameter {name} must be between {min_val} and {max_val}")
        
        super().set_parameter(name, value)
    
    def construct(self) -> str:
        return ""


class VLCEqualizerFilter(VLCAudioFilter):
    def __init__(self):
        vlc_param_map = {
            "preamp": ("preamp", float, (-20.0, 20.0)),
            "band_60": ("60Hz", float, (-20.0, 20.0)),
            "band_170": ("170Hz", float, (-20.0, 20.0)),
            "band_310": ("310Hz", float, (-20.0, 20.0)),
            "band_600": ("600Hz", float, (-20.0, 20.0)),
            "band_1000": ("1KHz", float, (-20.0, 20.0)),
            "band_3000": ("3KHz", float, (-20.0, 20.0)),
            "band_6000": ("6KHz", float, (-20.0, 20.0)),
            "band_12000": ("12KHz", float, (-20.0, 20.0)),
            "band_14000": ("14KHz", float, (-20.0, 20.0)),
            "band_16000": ("16KHz", float, (-20.0, 20.0))
        }
        
        parameters = {
            "preamp": 12.0,
            "band_60": 0.0,
            "band_170": 0.0,
            "band_310": 0.0,
            "band_600": 0.0,
            "band_1000": 0.0,
            "band_3000": 0.0,
            "band_6000": 0.0,
            "band_12000": 0.0,
            "band_14000": 0.0,
            "band_16000": 0.0
        }
        
        backend_info = {
            "vlc_filter_type": "equalizer",
            "vlc_param_map": vlc_param_map
        }
        
        super().__init__("Equalizer", parameters, backend_info)


import random
import os

from .media_info import MediaInfo
from .__AV_Interface import AVMediaInterface
from .__AV_Common import *


class AVMediaInstance:


    def __init__(self, controler:AVMediaInterface, file_path:str = ""):
        self.__controler = controler
        self.__media_type = self.__controler.type
        self.__media_backend = self.__controler.backend
        self.__media_source:AVMediaSource = AVMediaSource.AV_SRC_NOT_SET
        self.__file_path = file_path
        self.__id:int = random.randint(1000000, 9999999)
        self.__filters:dict[int, AVFilter] = {}


    @property
    def media_type(self):
        return self.__media_type

    @property
    def media_backend(self):
        return self.__media_backend

    @property
    def media_source(self):
        return self.__media_source

    @property
    def instance_id(self):
        return self.__id

    @property
    def file_path(self):
        return self.__file_path


    # Media control methods
    def load_file(self, file_path:str):
        #if self.get_playback_state() == AVPlaybackState.AV_STATE_PLAYING:
            #self.stop()
        if file_path != "" and is_path(file_path) and os.path.exists(file_path):
            self.__file_path = file_path
            self.__controler.load_file(self.__id, file_path)

    def load_url(self, url:str):
        if self.get_playback_state() == AVPlaybackState.AV_STATE_PLAYING:
            self.stop()
        if url != "" and is_url(url):
            self.__file_path = url
            self.__controler.load_url(self.__id, url)

    def play(self):
        self.__controler.play(self.__id)

    def pause(self):
        self.__controler.pause(self.__id)

    def stop(self):
        self.__controler.stop(self.__id)

    def mute(self):
        self.__controler.mute(self.__id)

    def unmute(self):
        self.__controler.unmute(self.__id)

    def set_position(self, position:int):
        self.__controler.set_position(self.__id, position)

    def set_volume(self, offset:float):
        self.__controler.set_volume(self.__id, offset)

    def set_loop(self, loop:bool):
        self.__controler.set_loop(self.__id, loop)

    def get_length(self) -> int:
        return self.__controler.get_length(self.__id, )

    def get_position(self) -> int:
        return self.__controler.get_position(self.__id, )

    def get_volume(self) -> float:
        return self.__controler.get_volume(self.__id, )

    def get_playback_state(self) -> AVPlaybackState:
        return self.__controler.get_play_state(self.__id, )

    def get_mute_state(self) -> AVMuteState:
        return self.__controler.get_mute_state(self.__id, )

    # Effects/filters
    def apply_filter(self, filter:AVFilter) -> int:
        if filter.type != self.media_type and filter.backend != self.media_backend:
            return -1
        filter_id = random.randint(1000000, 9999999)
        self.__filters[filter_id] = filter
        self.__controler.apply_filter(self.__id, filter_id, filter)
        return filter_id

    def remove_filter(self, filter_id:int):
        if filter_id in self.__filters:
            self.__filters.pop(filter_id)
            self.__controler.remove_filter(self.__id, filter_id )

    def remove_filters(self):

        for filter_id in list(self.__filters.keys()):
            try:
                self.remove_filter(filter_id)
            except:
                if filter_id in self.__filters:
                    self.__filters.pop(filter_id)

    def set_parameter(self, filter_id:int, parameter_name:str, parameter_value:ParameterValue):
        if filter_id in self.__filters and parameter_name in self.__filters[filter_id].get_parameters():
            self.__controler.set_parameter(self.__id, filter_id, parameter_name, parameter_value)

    def get_parameter(self, filter_id:int, parameter_name:str) -> ParameterValue | None:
        if filter_id in self.__filters and parameter_name in self.__filters[filter_id].get_parameters():
            return self.__controler.get_parameter(self.__id, filter_id, parameter_name)
        return None

    def release(self):

        try:
            if self.get_playback_state() == AVPlaybackState.AV_STATE_PLAYING:
                self.stop()
        except:
            pass
        

        try:
            self.remove_filters()
        except:
            pass
        

        try:
            self.__controler.release(self.__id)
        except:
            pass

    def media_info(self) -> Any:
        return MediaInfo(self.__file_path)



from typing import Any
from .__AV_Common import *
from .audio_filter import AudioFilter

class MPVAudioFilter(AVFilter):
    def __init__(self, filter_handle: str, parameters: dict[str, ParameterValue], backend_additional_info: dict[Any, Any]):
        super().__init__(AVFilterType.AV_TYPE_AUDIO, AVMediaBackend.AV_BACKEND_MPV, filter_handle, parameters, backend_additional_info)
        self._validate_parameters()
    
    def _validate_parameters(self):
        mpv_param_map = self.info.get("mpv_param_map", {})
        current_params = self.get_parameters()
        
        for param_name, param_value in current_params.items():
            if param_name in mpv_param_map:
                _, param_type, param_range = mpv_param_map[param_name]
                
                if not isinstance(param_value, param_type):
                    raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_TYPE, 
                                f"Parameter {param_name} must be of type {param_type.__name__}")
                
                if param_range and param_type in [int, float]:
                    min_val, max_val = param_range[:2]
                    if not (min_val <= param_value <= max_val):
                        raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_RANGE,
                                    f"Parameter {param_name} must be between {min_val} and {max_val}")
    
    def set_parameter(self, name: str, value: ParameterValue):
        mpv_param_map = self.info.get("mpv_param_map", {})
        if name in mpv_param_map:
            _, param_type, param_range = mpv_param_map[name]
            
            if not isinstance(value, param_type):
                raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_TYPE,
                            f"Parameter {name} must be of type {param_type.__name__}")
            
            if param_range and param_type in [int, float]:
                min_val, max_val = param_range[:2]
                if not (min_val <= value <= max_val):
                    raise AVError(AVErrorInfo.INVALID_FILTER_PARAMETER_RANGE,
                                f"Parameter {name} must be between {min_val} and {max_val}")
        
        super().set_parameter(name, value)
    
    def construct(self) -> str:
        mpv_filter_name = self.info.get("mpv_filter_name", "")
        effect_syntax = self.info.get("effect_syntax", "lavfi")
        mpv_param_map = self.info.get("mpv_param_map", {})
        
        if not mpv_filter_name:
            return ""
        
        params = self.get_parameters()
        param_strings = []
        
        for param_name, param_value in params.items():
            if param_name in mpv_param_map:
                mpv_param_name, _, _ = mpv_param_map[param_name]
                param_strings.append(f"{mpv_param_name}={param_value}")
        
        param_string = ":".join(param_strings)
        
        if effect_syntax == "lavfi":
            if param_string:
                return f"lavfi=[{mpv_filter_name}={param_string}]"
            else:
                return f"lavfi=[{mpv_filter_name}]"
        elif effect_syntax == "@rb":
            if param_string:
                return f"@rb:{mpv_filter_name}={param_string}"
            else:
                return f"@rb:{mpv_filter_name}"
        else:
            if param_string:
                return f"{mpv_filter_name}={param_string}"
            else:
                return mpv_filter_name


class MPVEchoFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "in_gain": ("in_gain", float, (0.0, 1.0, 0.05, 0.6)),
            "out_gain": ("out_gain", float, (0.0, 1.0, 0.05, 0.3)),
            "delays": ("delays", str, ()),
            "decays": ("decays", str, ())
        }
        
        parameters = {
            "in_gain": 0.6,
            "out_gain": 0.3,
            "delays": "1000",
            "decays": "0.5"
        }
        
        backend_info = {
            "mpv_filter_name": "aecho",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Echo", parameters, backend_info)

class MPVReverbFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "dry": ("dry", float, (0.0, 1.0, 0.1, 1.0)),
            "wet": ("wet", float, (0.0, 1.0, 0.1, 0.3)),
            "length": ("length", int, (1, 100, 1, 1)),
            "irnorm": ("irnorm", float, (-1.0, 2.0, 0.1, 1.0)),
            "irgain": ("irgain", float, (0.0, 1.0, 0.1, 1.0))
        }
        
        parameters = {
            "dry": 1.0,
            "wet": 0.3,
            "length": 1,
            "irnorm": 1.0,
            "irgain": 1.0
        }
        
        backend_info = {
            "mpv_filter_name": "afir",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Reverb", parameters, backend_info)

class MPVLowPassFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "frequency": ("frequency", float, (20.0, 20000.0, 10.0, 500.0)),
            "poles": ("poles", int, (1, 2, 1, 2)),
            "width": ("width", float, (0.1, 10.0, 0.1, 0.707)),
            "mix": ("mix", float, (0.0, 1.0, 0.1, 1.0))
        }
        
        parameters = {
            "frequency": 500.0,
            "poles": 2,
            "width": 0.707,
            "mix": 1.0
        }
        
        backend_info = {
            "mpv_filter_name": "lowpass",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Low Pass", parameters, backend_info)

class MPVHighPassFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "frequency": ("frequency", float, (20.0, 20000.0, 10.0, 3000.0)),
            "poles": ("poles", int, (1, 2, 1, 2)),
            "width": ("width", float, (0.1, 10.0, 0.1, 0.707)),
            "mix": ("mix", float, (0.0, 1.0, 0.1, 1.0))
        }
        
        parameters = {
            "frequency": 3000.0,
            "poles": 2,
            "width": 0.707,
            "mix": 1.0
        }
        
        backend_info = {
            "mpv_filter_name": "highpass",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("High Pass", parameters, backend_info)

class MPVCompressorFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "level_in": ("level_in", float, (0.015625, 64.0, 0.1, 1.0)),
            "threshold": ("threshold", float, (0.00097563, 1.0, 0.01, 0.125)),
            "ratio": ("ratio", float, (1.0, 20.0, 0.1, 2.0)),
            "attack": ("attack", float, (0.01, 2000.0, 1.0, 20.0)),
            "release": ("release", float, (0.01, 9000.0, 1.0, 250.0)),
            "makeup": ("makeup", float, (1.0, 64.0, 0.1, 1.0)),
            "knee": ("knee", float, (1.0, 8.0, 0.1, 2.82843)),
            "detection": ("detection", str, ()),
            "mix": ("mix", float, (0.0, 1.0, 0.1, 1.0))
        }
        
        parameters = {
            "level_in": 1.0,
            "threshold": 0.125,
            "ratio": 2.0,
            "attack": 20.0,
            "release": 250.0,
            "makeup": 1.0,
            "knee": 2.82843,
            "detection": "rms",
            "mix": 1.0
        }
        
        backend_info = {
            "mpv_filter_name": "acompressor",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Compressor", parameters, backend_info)

class MPVFlangerFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "delay": ("delay", float, (0.0, 30.0, 0.1, 0.0)),
            "depth": ("depth", float, (0.0, 10.0, 0.1, 2.0)),
            "regen": ("regen", float, (-95.0, 95.0, 1.0, 0.0)),
            "width": ("width", float, (0.0, 100.0, 1.0, 71.0)),
            "speed": ("speed", float, (0.1, 10.0, 0.1, 0.5)),
            "shape": ("shape", str, ()),
            "phase": ("phase", float, (0.0, 100.0, 1.0, 25.0)),
            "interp": ("interp", str, ())
        }
        
        parameters = {
            "delay": 0.0,
            "depth": 2.0,
            "regen": 0.0,
            "width": 71.0,
            "speed": 0.5,
            "shape": "sinusoidal",
            "phase": 25.0,
            "interp": "linear"
        }
        
        backend_info = {
            "mpv_filter_name": "flanger",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Flanger", parameters, backend_info)

class MPVChorusFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "in_gain": ("in_gain", float, (0.0, 1.0, 0.05, 0.4)),
            "out_gain": ("out_gain", float, (0.0, 1.0, 0.05, 0.4)),
            "delays": ("delays", str, ()),
            "decays": ("decays", str, ()),
            "speeds": ("speeds", str, ()),
            "depths": ("depths", str, ())
        }
        
        parameters = {
            "in_gain": 0.4,
            "out_gain": 0.4,
            "delays": "55",
            "decays": "0.4",
            "speeds": "0.25",
            "depths": "2"
        }
        
        backend_info = {
            "mpv_filter_name": "chorus",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Chorus", parameters, backend_info)

class MPVPitchShiftFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "pitch-scale": ("pitch-scale", float, (1.0, 100.0, 1.0, 1.0)),
            "engine": ("engine", str, ("faster", "finer"))
        }
        
        parameters = {
            "pitch-scale": 1.0,
            "engine": "finer"
        }
        
        backend_info = {
            "mpv_filter_name": "rubberband",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "@rb:"
        }
        
        super().__init__("Pitch Shift", parameters, backend_info)

class MPVTempoScaleFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "scale": ("scale", int, (1, 25, 1, 1)),
            "speed": ("speed", str, ("pitch", "none"))
        }
        
        parameters = {
            "scale": 1,
            "speed": "none"
        }
        
        backend_info = {
            "mpv_filter_name": "scaletempo",
            "mpv_param_map": mpv_param_map
        }
        
        super().__init__("Tempo Scale", parameters, backend_info)

class MPVLimiterFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "level_in": ("level_in", float, (0.0, 64.0, 0.1, 1.0)),
            "level_out": ("level_out", float, (0.0, 64.0, 0.1, 1.0)),
            "limit": ("limit", float, (0.0, 1.0, 0.01, 1.0)),
            "attack": ("attack", float, (0.1, 1000.0, 0.1, 5.0)),
            "release": ("release", float, (1.0, 9000.0, 1.0, 50.0)),
            "asc": ("asc", str, ()),
            "asc_level": ("asc_level", float, (0.0, 1.0, 0.1, 0.5)),
            "level": ("level", str, ())
        }
        
        parameters = {
            "level_in": 1.0,
            "level_out": 1.0,
            "limit": 1.0,
            "attack": 5.0,
            "release": 50.0,
            "asc": "false",
            "asc_level": 0.5,
            "level": "true"
        }
        
        backend_info = {
            "mpv_filter_name": "alimiter",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Limiter", parameters, backend_info)

class MPVBandPassFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "frequency": ("frequency", float, (20.0, 20000.0, 10.0, 3000.0)),
            "width": ("width", float, (0.1, 1000.0, 0.1, 100.0)),
            "csg": ("csg", int, (0, 1, 1, 0)),
            "mix": ("mix", float, (0.0, 1.0, 0.1, 1.0)),
            "width_type": ("width_type", str, ())
        }
        
        parameters = {
            "frequency": 3000.0,
            "width": 100.0,
            "csg": 0,
            "mix": 1.0,
            "width_type": "h"
        }
        
        backend_info = {
            "mpv_filter_name": "bandpass",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Band Pass", parameters, backend_info)

class MPVGateFilter(MPVAudioFilter):
    def __init__(self):
        mpv_param_map = {
            "level_in": ("level_in", float, (0.015625, 64.0, 0.1, 1.0)),
            "mode": ("mode", str, ()),
            "range": ("range", float, (0.0, 1.0, 0.01, 0.06125)),
            "threshold": ("threshold", float, (0.0, 1.0, 0.01, 0.125)),
            "ratio": ("ratio", float, (1.0, 9000.0, 0.1, 2.0)),
            "attack": ("attack", float, (0.01, 9000.0, 1.0, 20.0)),
            "release": ("release", float, (0.01, 9000.0, 1.0, 250.0)),
            "makeup": ("makeup", float, (1.0, 64.0, 0.1, 1.0)),
            "knee": ("knee", float, (1.0, 8.0, 0.1, 2.828427125)),
            "detection": ("detection", str, ()),
            "link": ("link", str, ())
        }
        
        parameters = {
            "level_in": 1.0,
            "mode": "downward",
            "range": 0.06125,
            "threshold": 0.125,
            "ratio": 2.0,
            "attack": 20.0,
            "release": 250.0,
            "makeup": 1.0,
            "knee": 2.828427125,
            "detection": "rms",
            "link": "average"
        }
        
        backend_info = {
            "mpv_filter_name": "agate",
            "mpv_param_map": mpv_param_map,
            "effect_syntax": "lavfi"
        }
        
        super().__init__("Gate", parameters, backend_info)
        